import argparse
import os
import yaml

from jsonschema import Draft4Validator


def parse_args() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="""
This script validates netplan example files against the cloud-init networkv2
schema.  The goal is to provide an easy way to track drift between networkv2
and netplan.

Netplan does not provide a centralized schema file to compare against directly,
but does maintain a reasonably comprehensive examples directory.  This script
relies on that directory to glean the netplan schema.
            """,
        formatter_class=argparse.RawTextHelpFormatter,
    )
    parser.add_argument(
        "--schema-file",
        required=True,
        help="""
The cloud-init networkv2 schema file found in
cloudinit/config/schemas/schema-network-config-v2.json
in this repository.
        """,
    )
    parser.add_argument(
        "--netplan-examples",
        required=True,
        help="""
The examples/ directory in the netplan repo.  The most
recent netplan repo can be cloned from
https://github.com/canonical/netplan.
        """,
    )
    return parser.parse_args()


def validate_netplan_against_networkv2(schema_file, netplan_examples):
    """
    This script validates netplan example files against the cloud-init
    networkv2 schema.  The goal is to provide an easy way to track drift
    between networkv2 and netplan.

    Netplan does not provide a centralized schema file to compare against
    directly, but does maintain a reasonably comprehensive examples directory.
    This script relies on that directory to glean the netplan schema.

    There are two arguments required for this script.  The first is schema_file
    which should point to the cloud-init networkv2 schema file found in
    cloudinit/config/schemas/schema-network-config-v2.json in this repository.
    The second is netplan_examples which should point to the examples/
    directory in the netplan repo.  The most recent netplan repo can be cloned
    from https://github.com/canonical/netplan.
    """
    networkv2_schema = None
    with open(schema_file) as f:
        networkv2_schema = yaml.safe_load(f)
    validator = Draft4Validator(networkv2_schema)

    error_obj = {}
    for walk_tuple in os.walk(netplan_examples):
        filenames = walk_tuple[2]
        for fname in filenames:
            if fname.endswith(".yaml"):
                with open(netplan_examples + fname) as netplan_f:
                    netplan_example = yaml.safe_load(netplan_f)
                    errors = validator.iter_errors(netplan_example)
                    for e in errors:
                        schema_path_str = "-".join(map(str, e.schema_path))
                        if schema_path_str not in error_obj:
                            error_obj[schema_path_str] = {}
                        if e.message not in error_obj[schema_path_str]:
                            error_obj[schema_path_str][e.message] = set()
                        error_obj[schema_path_str][e.message].add(fname)

    # clean up error_obj for human readability
    for schema_path in error_obj:
        for message in error_obj[schema_path]:
            error_obj[schema_path][message] = list(
                error_obj[schema_path][message]
            )
    print(yaml.dump(error_obj))


if __name__ == "__main__":
    args = parse_args()
    validate_netplan_against_networkv2(args.schema_file, args.netplan_examples)



import uuid
from typing import cast

import pytest
from pycloudlib.lxd.instance import LXDInstance

from cloudinit import subp
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import PLATFORM
from tests.integration_tests.releases import CURRENT_RELEASE, FOCAL

_INSTANCE_ID = uuid.uuid4()


def setup_meta_data(instance: LXDInstance):
    """Increment the instance id and apply it to the instance."""
    global _INSTANCE_ID
    _INSTANCE_ID = uuid.UUID(int=(1 + _INSTANCE_ID.int))
    command = [
        "lxc",
        "config",
        "set",
        instance.name,
        f"volatile.cloud-init.instance-id={_INSTANCE_ID}",
    ]
    assert subp.subp(command)


# class TestInstanceID:
@pytest.mark.skipif(
    PLATFORM not in ["lxd_container", "lxd_vm"] or CURRENT_RELEASE == FOCAL,
    reason="Uses lxd-specific behavior.",
)
@pytest.mark.lxd_setup.with_args(setup_meta_data)
@pytest.mark.lxd_use_exec
def test_instance_id_changes(client: IntegrationInstance):
    """Verify instance id change behavior

    If the id from the datasource changes, cloud-init should update the
    instance id link.
    """
    client.execute("cloud-init status --wait")
    # check that instance id is the one we set
    assert (
        str(_INSTANCE_ID)
        == client.execute("cloud-init query instance-id").stdout.rstrip()
    )
    assert (
        f"/var/lib/cloud/instances/{_INSTANCE_ID}"
        == client.execute(
            "readlink -f /var/lib/cloud/instance"
        ).stdout.rstrip()
    )

    instance = cast(LXDInstance, client.instance)
    setup_meta_data(instance)
    client.restart()
    client.execute("cloud-init status --wait")
    # check that instance id is the one we reset
    assert (
        str(_INSTANCE_ID)
        == client.execute("cloud-init query instance-id").stdout.rstrip()
    )
    assert (
        f"/var/lib/cloud/instances/{_INSTANCE_ID}"
        == client.execute(
            "readlink -f /var/lib/cloud/instance"
        ).stdout.rstrip()
    )


@pytest.mark.lxd_use_exec
def test_instance_id_no_changes(client: IntegrationInstance):
    """Verify instance id no change behavior

    If the id from the datasource does not change, cloud-init should not
    update the instance id link.
    """
    instance_id = client.execute(
        "cloud-init query instance-id"
    ).stdout.rstrip()
    assert (
        f"/var/lib/cloud/instances/{instance_id}"
        == client.execute(
            "readlink -f /var/lib/cloud/instance"
        ).stdout.rstrip()
    )
    client.restart()
    client.execute("cloud-init status --wait")
    assert (
        instance_id
        == client.execute("cloud-init query instance-id").stdout.rstrip()
    )
    assert (
        f"/var/lib/cloud/instances/{instance_id}"
        == client.execute(
            "readlink -f /var/lib/cloud/instance"
        ).stdout.rstrip()
    )



# Copyright (C) 2011 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""Final Message: Output final message when cloud-init has finished"""

import logging

from cloudinit import templater, util, version
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.distros import ALL_DISTROS
from cloudinit.log import log_util
from cloudinit.settings import PER_ALWAYS

frequency = PER_ALWAYS
meta: MetaSchema = {
    "id": "cc_final_message",
    "distros": [ALL_DISTROS],
    "frequency": frequency,
    "activate_by_schema_keys": [],
}

LOG = logging.getLogger(__name__)

# Jinja formatted default message
FINAL_MESSAGE_DEF = (
    "## template: jinja\n"
    "Cloud-init v. {{version}} finished at {{timestamp}}."
    " Datasource {{datasource}}.  Up {{uptime}} seconds"
)


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:

    msg_in = ""
    if args:
        msg_in = str(args[0])
    else:
        msg_in = util.get_cfg_option_str(cfg, "final_message", "")

    msg_in = msg_in.strip()
    if not msg_in:
        msg_in = FINAL_MESSAGE_DEF

    uptime = util.uptime()
    ts = util.time_rfc2822()
    cver = version.version_string()
    try:
        subs = {
            "uptime": uptime,
            "timestamp": ts,
            "version": cver,
            "datasource": str(cloud.datasource),
        }
        subs.update(dict([(k.upper(), v) for k, v in subs.items()]))
        log_util.multi_log(
            "%s\n" % (templater.render_string(msg_in, subs)),
            console=False,
            stderr=True,
            log=LOG,
        )
    except templater.JinjaSyntaxParsingException as e:
        util.logexc(
            LOG, "Failed to render templated final message: %s", str(e)
        )
    except Exception:
        util.logexc(LOG, "Failed to render final message template")

    boot_fin_fn = cloud.paths.boot_finished
    try:
        contents = "%s - %s - v. %s\n" % (uptime, ts, cver)
        util.write_file(boot_fin_fn, contents, ensure_dir_exists=False)
    except Exception:
        util.logexc(LOG, "Failed to write boot finished file %s", boot_fin_fn)

    if cloud.datasource.dsname == "None":
        if cloud.datasource.sys_cfg.get("datasource_list") != ["None"]:
            LOG.warning("Used fallback datasource")



# Copyright (C) 2017 Canonical Ltd.
#
# This file is part of cloud-init. See LICENSE file for license information.

import argparse
import re
import sys
from datetime import datetime, timezone
from typing import IO, Dict, List, Optional, Tuple, Union

from cloudinit.analyze import dump, show
from cloudinit.atomic_helper import json_dumps


def get_parser(
    parser: Optional[argparse.ArgumentParser] = None,
) -> argparse.ArgumentParser:
    if not parser:
        parser = argparse.ArgumentParser(
            prog="cloudinit-analyze",
            description="Devel tool: Analyze cloud-init logs and data",
        )
    subparsers = parser.add_subparsers(title="Subcommands", dest="subcommand")
    subparsers.required = True

    parser_blame = subparsers.add_parser(
        "blame", help="Print list of executed stages ordered by time to init"
    )
    parser_blame.add_argument(
        "-i",
        "--infile",
        action="store",
        dest="infile",
        default="/var/log/cloud-init.log",
        help="specify where to read input.",
    )
    parser_blame.add_argument(
        "-o",
        "--outfile",
        action="store",
        dest="outfile",
        default="-",
        help="specify where to write output. ",
    )
    parser_blame.set_defaults(action=("blame", analyze_blame))

    parser_show = subparsers.add_parser(
        "show", help="Print list of in-order events during execution"
    )
    parser_show.add_argument(
        "-f",
        "--format",
        action="store",
        dest="print_format",
        default="%I%D @%Es +%ds",
        help="specify formatting of output.",
    )
    parser_show.add_argument(
        "-i",
        "--infile",
        action="store",
        dest="infile",
        default="/var/log/cloud-init.log",
        help="specify where to read input.",
    )
    parser_show.add_argument(
        "-o",
        "--outfile",
        action="store",
        dest="outfile",
        default="-",
        help="specify where to write output.",
    )
    parser_show.set_defaults(action=("show", analyze_show))
    parser_dump = subparsers.add_parser(
        "dump", help="Dump cloud-init events in JSON format"
    )
    parser_dump.add_argument(
        "-i",
        "--infile",
        action="store",
        dest="infile",
        default="/var/log/cloud-init.log",
        help="specify where to read input. ",
    )
    parser_dump.add_argument(
        "-o",
        "--outfile",
        action="store",
        dest="outfile",
        default="-",
        help="specify where to write output. ",
    )
    parser_dump.set_defaults(action=("dump", analyze_dump))
    parser_boot = subparsers.add_parser(
        "boot", help="Print list of boot times for kernel and cloud-init"
    )
    parser_boot.add_argument(
        "-i",
        "--infile",
        action="store",
        dest="infile",
        default="/var/log/cloud-init.log",
        help="specify where to read input. ",
    )
    parser_boot.add_argument(
        "-o",
        "--outfile",
        action="store",
        dest="outfile",
        default="-",
        help="specify where to write output.",
    )
    parser_boot.set_defaults(action=("boot", analyze_boot))
    return parser


def analyze_boot(name: str, args: argparse.Namespace) -> str:
    """Report a list of how long different boot operations took.

    For Example:
    -- Most Recent Boot Record --
        Kernel Started at: <time>
        Kernel ended boot at: <time>
        Kernel time to boot (seconds): <time>
        Cloud-init activated by systemd at: <time>
        Time between Kernel end boot and Cloud-init activation (seconds):<time>
        Cloud-init start: <time>
    """
    infh, outfh = configure_io(args)
    kernel_info = show.dist_check_timestamp()
    status_code, kernel_start, kernel_end, ci_sysd_start = kernel_info
    kernel_start_timestamp = datetime.fromtimestamp(kernel_start, timezone.utc)
    kernel_end_timestamp = datetime.fromtimestamp(kernel_end, timezone.utc)
    ci_sysd_start_timestamp = datetime.fromtimestamp(
        ci_sysd_start, timezone.utc
    )
    try:
        last_init_local = [
            e
            for e in _get_events(infh)
            if e["name"] == "init-local"
            and "starting search" in str(e["description"])
        ][-1]
        ci_start: Union[datetime, str] = datetime.fromtimestamp(
            float(last_init_local["timestamp"]), timezone.utc
        )
    except IndexError:
        ci_start = "Could not find init-local log-line in cloud-init.log"
        status_code = show.FAIL_CODE

    FAILURE_MSG = (
        "Your Linux distro or container does not support this "
        "functionality.\n"
        "You must be running a Kernel Telemetry supported "
        "distro.\nPlease check "
        "https://docs.cloud-init.io/en/latest"
        "/topics/analyze.html for more "
        "information on supported distros.\n"
    )

    SUCCESS_MSG = (
        "-- Most Recent Boot Record --\n"
        "    Kernel Started at: {k_s_t}\n"
        "    Kernel ended boot at: {k_e_t}\n"
        "    Kernel time to boot (seconds): {k_r}\n"
        "    Cloud-init activated by systemd at: {ci_sysd_t}\n"
        "    Time between Kernel end boot and Cloud-init "
        "activation (seconds): {bt_r}\n"
        "    Cloud-init start: {ci_start}\n"
    )

    CONTAINER_MSG = (
        "-- Most Recent Container Boot Record --\n"
        "    Container started at: {k_s_t}\n"
        "    Cloud-init activated by systemd at: {ci_sysd_t}\n"
        "    Cloud-init start: {ci_start}\n"
    )
    status_map = {
        show.FAIL_CODE: FAILURE_MSG,
        show.CONTAINER_CODE: CONTAINER_MSG,
        show.SUCCESS_CODE: SUCCESS_MSG,
    }

    kernel_runtime = kernel_end - kernel_start
    between_process_runtime = ci_sysd_start - kernel_end

    kwargs = {
        "k_s_t": kernel_start_timestamp,
        "k_e_t": kernel_end_timestamp,
        "k_r": kernel_runtime,
        "bt_r": between_process_runtime,
        "k_e": kernel_end,
        "k_s": kernel_start,
        "ci_sysd": ci_sysd_start,
        "ci_sysd_t": ci_sysd_start_timestamp,
        "ci_start": ci_start,
    }

    outfh.write(status_map[status_code].format(**kwargs))
    clean_io(infh, outfh)
    return status_code


def analyze_blame(name, args: argparse.Namespace) -> None:
    """Report a list of records sorted by largest time delta.

    For example:
      30.210s (init-local) searching for datasource
       8.706s (init-network) reading and applying user-data
        166ms (modules-config) ....
        807us (modules-final) ...

    We generate event records parsing cloud-init logs, formatting the output
    and sorting by record data ('delta')
    """
    infh, outfh = configure_io(args)
    blame_format = "     %ds (%n)"
    r = re.compile(r"(^\s+\d+\.\d+)", re.MULTILINE)
    idx = 0
    for idx, record in enumerate(
        show.show_events(_get_events(infh), blame_format)
    ):
        srecs = sorted(filter(r.match, record), reverse=True)
        outfh.write("-- Boot Record %02d --\n" % (idx + 1))
        outfh.write("\n".join(srecs) + "\n")
        outfh.write("\n")
    outfh.write("%d boot records analyzed\n" % (idx + 1))
    clean_io(infh, outfh)


def analyze_show(name, args: argparse.Namespace) -> None:
    """Generate output records using the 'standard' format to printing events.

    Example output follows:
        Starting stage: (init-local)
          ...
        Finished stage: (init-local) 0.105195 seconds

        Starting stage: (init-network)
          ...
        Finished stage: (init-network) 0.339024 seconds

        Starting stage: (modules-config)
          ...
        Finished stage: (modules-config) 0.NNN seconds

        Starting stage: (modules-final)
          ...
        Finished stage: (modules-final) 0.NNN seconds
    """
    infh, outfh = configure_io(args)
    for idx, record in enumerate(
        show.show_events(_get_events(infh), args.print_format)
    ):
        outfh.write("-- Boot Record %02d --\n" % (idx + 1))
        outfh.write(
            "The total time elapsed since completing an event is"
            ' printed after the "@" character.\n'
        )
        outfh.write(
            'The time the event takes is printed after the "+" character.\n\n'
        )
        outfh.write("\n".join(record) + "\n")
    outfh.write("%d boot records analyzed\n" % (idx + 1))
    clean_io(infh, outfh)


def analyze_dump(name, args: argparse.Namespace) -> None:
    """Dump cloud-init events in json format"""
    infh, outfh = configure_io(args)
    outfh.write(json_dumps(_get_events(infh)) + "\n")
    clean_io(infh, outfh)


def _get_events(infile: IO) -> List[Dict[str, Union[str, float]]]:
    rawdata = None
    events, rawdata = show.load_events_infile(infile)
    if not events:
        events, _ = dump.dump_events(rawdata=rawdata)
    return events


def configure_io(args: argparse.Namespace) -> Tuple[IO, IO]:
    """Common parsing and setup of input/output files"""
    if args.infile == "-":
        infh = sys.stdin
    else:
        try:
            infh = open(args.infile, "r")
        except OSError:
            sys.stderr.write("Cannot open file %s\n" % args.infile)
            sys.exit(1)

    if args.outfile == "-":
        outfh = sys.stdout
    else:
        try:
            outfh = open(args.outfile, "w")
        except OSError:
            sys.stderr.write("Cannot open file %s\n" % args.outfile)
            sys.exit(1)

    return infh, outfh


def clean_io(*file_handles: IO) -> None:
    """close filehandles"""
    for file_handle in file_handles:
        if file_handle in (sys.stdin, sys.stdout):
            continue
        file_handle.close()


if __name__ == "__main__":
    parser = get_parser()
    args = parser.parse_args()
    (name, action_functor) = args.action
    action_functor(name, args)



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import collections.abc  # pylint: disable=import-error
import copy
import io
import logging
import logging.config
import logging.handlers
import os
import sys
import time
from collections import defaultdict
from contextlib import suppress
from typing import DefaultDict

DEFAULT_LOG_FORMAT = "%(asctime)s - %(filename)s[%(levelname)s]: %(message)s"
DEPRECATED = 35
TRACE = logging.DEBUG - 5


class CustomLoggerType(logging.Logger):
    """A hack to get mypy to stop complaining about custom logging methods.

    When using deprecated or trace logging, rather than:
        LOG = logging.getLogger(__name__)
    Instead do:
        LOG = cast(CustomLoggerType, logging.getLogger(__name__))
    """

    def trace(self, *args, **kwargs):
        pass

    def deprecated(self, *args, **kwargs):
        pass


def setup_basic_logging(level=logging.DEBUG, formatter=None):
    formatter = formatter or logging.Formatter(DEFAULT_LOG_FORMAT)
    root = logging.getLogger()
    console = logging.StreamHandler(sys.stderr)
    console.setFormatter(formatter)
    console.setLevel(level)
    root.addHandler(console)
    root.setLevel(level)


def flush_loggers(root):
    if not root:
        return
    for h in root.handlers:
        if isinstance(h, (logging.StreamHandler)):
            with suppress(IOError):
                h.flush()
    flush_loggers(root.parent)


def define_extra_loggers() -> None:
    """Add DEPRECATED and TRACE log levels to the logging module."""

    def new_logger(level):
        def log_at_level(self, message, *args, **kwargs):
            if self.isEnabledFor(level):
                self._log(level, message, args, **kwargs)

        return log_at_level

    logging.addLevelName(DEPRECATED, "DEPRECATED")
    logging.addLevelName(TRACE, "TRACE")
    setattr(logging.Logger, "deprecated", new_logger(DEPRECATED))
    setattr(logging.Logger, "trace", new_logger(TRACE))


def setup_logging(cfg=None):
    # See if the config provides any logging conf...
    if not cfg:
        cfg = {}

    root_logger = logging.getLogger()
    exporter = LogExporter()
    exporter.setLevel(logging.WARN)

    log_cfgs = []
    log_cfg = cfg.get("logcfg")
    if log_cfg and isinstance(log_cfg, str):
        # If there is a 'logcfg' entry in the config,
        # respect it, it is the old keyname
        log_cfgs.append(str(log_cfg))
    elif "log_cfgs" in cfg:
        for a_cfg in cfg["log_cfgs"]:
            if isinstance(a_cfg, str):
                log_cfgs.append(a_cfg)
            elif isinstance(a_cfg, (collections.abc.Iterable)):
                cfg_str = [str(c) for c in a_cfg]
                log_cfgs.append("\n".join(cfg_str))
            else:
                log_cfgs.append(str(a_cfg))

    # See if any of them actually load...
    am_tried = 0

    # log_cfg may contain either a filepath to a file containing a logger
    # configuration, or a string containing a logger configuration
    # https://docs.python.org/3/library/logging.config.html#logging-config-fileformat
    for log_cfg in log_cfgs:
        # The default configuration includes an attempt at using /dev/log,
        # followed up by writing to a file. /dev/log will not exist in
        # very early boot, so an exception on that is expected.
        with suppress(FileNotFoundError):
            am_tried += 1

            # If the value is not a filename, assume that it is a config.
            if not (log_cfg.startswith("/") and os.path.isfile(log_cfg)):
                log_cfg = io.StringIO(log_cfg)

            # Attempt to load its config.
            logging.config.fileConfig(log_cfg)

            # Configure warning exporter after loading logging configuration
            root_logger.addHandler(exporter)

            # Use the first valid configuration.
            return

    # Configure warning exporter for basic logging
    root_logger.addHandler(exporter)

    # If it didn't work, at least setup a basic logger (if desired)
    basic_enabled = cfg.get("log_basic", True)

    sys.stderr.write(
        "WARN: no logging configured! (tried %s configs)\n" % (am_tried)
    )
    if basic_enabled:
        sys.stderr.write("Setting up basic logging...\n")
        setup_basic_logging()


class LogExporter(logging.StreamHandler):
    holder: DefaultDict[str, list] = defaultdict(list)

    def emit(self, record: logging.LogRecord):
        self.holder[record.levelname].append(record.getMessage())

    def export_logs(self):
        return copy.deepcopy(self.holder)

    def clean_logs(self):
        self.holder = defaultdict(list)

    def flush(self):
        pass


def reset_logging():
    """Remove all current handlers and unset log level."""
    log = logging.getLogger()
    handlers = list(log.handlers)
    for h in handlers:
        h.flush()
        h.close()
        log.removeHandler(h)
    log.setLevel(logging.NOTSET)


def setup_backup_logging():
    """In the event that internal logging exception occurs and logging is not
    possible for some reason, make a desperate final attempt to log to stderr
    which may ease debugging.
    """
    fallback_handler = logging.StreamHandler(sys.stderr)
    setattr(fallback_handler, "handleError", lambda record: None)
    fallback_handler.setFormatter(
        logging.Formatter(
            "FALLBACK: %(asctime)s - %(filename)s[%(levelname)s]: %(message)s"
        )
    )

    def handleError(self, record):
        """A closure that emits logs on stderr when other methods fail"""
        with suppress(IOError):
            fallback_handler.handle(record)
            fallback_handler.flush()

    setattr(logging.Handler, "handleError", handleError)


class CloudInitLogRecord(logging.LogRecord):
    """reporting the filename as __init__.py isn't very useful in logs

    if the filename is __init__.py, use the parent directory as the filename
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if "__init__.py" == self.filename:
            self.filename = os.path.basename(os.path.dirname(self.pathname))


def configure_root_logger():
    """Customize the root logger for cloud-init"""

    # Always format logging timestamps as UTC time
    logging.Formatter.converter = time.gmtime
    define_extra_loggers()
    setup_backup_logging()
    reset_logging()

    # add handler only to the root logger
    handler = LogExporter()
    handler.setLevel(logging.WARN)
    logging.getLogger().addHandler(handler)

    # LogRecord allows us to report more useful information than __init__.py
    logging.setLogRecordFactory(CloudInitLogRecord)



import os
from copy import deepcopy
from typing import Optional, Sequence, Set

from hypothesis import given, settings
from hypothesis_jsonschema import from_schema

from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)

settings.register_profile("ci", max_examples=1000)
settings.load_profile(os.getenv("HYPOTHESIS_PROFILE", "default"))


def remove_modules(schema, modules: Set[str]) -> dict:
    indices_to_delete = set()
    for module in set(modules):
        for index, ref_dict in enumerate(schema["allOf"]):
            if ref_dict["$ref"] == f"#/$defs/{module}":
                indices_to_delete.add(index)
                continue  # module found
    for index in indices_to_delete:
        schema["allOf"].pop(index)
    return schema


def remove_defs(schema, defs: Set[str]) -> dict:
    defs_to_delete = set(schema["$defs"].keys()).intersection(set(defs))
    for key in defs_to_delete:
        del schema["$defs"][key]
    return schema


def clean_schema(
    schema=None,
    modules: Optional[Sequence[str]] = None,
    defs: Optional[Sequence[str]] = None,
):
    schema = deepcopy(schema or get_schema())
    if modules:
        remove_modules(schema, set(modules))
    if defs:
        remove_defs(schema, set(defs))
    del schema["properties"]
    del schema["additionalProperties"]
    return schema


class TestSchemaFuzz:
    # Avoid https://github.com/Zac-HD/hypothesis-jsonschema/issues/97
    SCHEMA = clean_schema(
        modules=["cc_users_groups"],
        defs=["users_groups.groups_by_groupname", "users_groups.user"],
    )

    @given(from_schema(SCHEMA))
    def test_validate_full_schema(self, orig_config):
        config = deepcopy(orig_config)
        valid_props = get_schema()["properties"].keys()
        for key in orig_config.keys():
            if key not in valid_props:
                del config[key]
        try:
            validate_cloudconfig_schema(config, strict=True)
        except SchemaValidationError as ex:
            if ex.has_errors():
                raise



# Copyright (C) 2012 Canonical Ltd.
# Copyright (C) 2012 Hewlett-Packard Development Company, L.P.
# Copyright (C) 2012 Yahoo! Inc.
#
# Author: Scott Moser <scott.moser@canonical.com>
# Author: Juerg Haefliger <juerg.haefliger@hp.com>
# Author: Joshua Harlow <harlowja@yahoo-inc.com>
# Author: Ben Howard <ben.howard@canonical.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import copy
import logging

from cloudinit.distros import PREFERRED_NTP_CLIENTS, debian
from cloudinit.distros.package_management.snap import Snap
from cloudinit.net import activators
from cloudinit.net.netplan import CLOUDINIT_NETPLAN_FILE

LOG = logging.getLogger(__name__)


class Distro(debian.Distro):
    def __init__(self, name, cfg, paths):
        super().__init__(name, cfg, paths)
        # Ubuntu specific network cfg locations
        self.network_conf_fn = {
            "eni": "/etc/network/interfaces.d/50-cloud-init.cfg",
            "netplan": CLOUDINIT_NETPLAN_FILE,
        }
        self.renderer_configs = {
            "eni": {
                "eni_path": self.network_conf_fn["eni"],
                "eni_header": debian.NETWORK_FILE_HEADER,
            },
            "netplan": {
                "netplan_path": self.network_conf_fn["netplan"],
                "netplan_header": debian.NETWORK_FILE_HEADER,
                "postcmds": True,
            },
        }
        self.snap = Snap(self._runner)
        self.package_managers.append(self.snap)

    def package_command(self, command, args=None, pkgs=None):
        super().package_command(command, args, pkgs)
        if self.snap.available():
            self.snap.upgrade_packages()

    @property
    def preferred_ntp_clients(self):
        """The preferred ntp client is dependent on the version."""
        if not self._preferred_ntp_clients:
            self._preferred_ntp_clients = copy.deepcopy(PREFERRED_NTP_CLIENTS)
        return self._preferred_ntp_clients

    def wait_for_network(self) -> None:
        """Ensure that cloud-init's network service has network connectivity"""
        try:
            self.network_activator.wait_for_network()
        except activators.NoActivatorException:
            LOG.error("Failed to wait for network. No network activator found")
        except Exception as e:
            LOG.error("Failed to wait for network: %s", e)



# This file is part of cloud-init. See LICENSE file for license information.
"""Series of integration tests covering apt functionality."""
import logging
import re
from textwrap import dedent

import pytest

from cloudinit.config import cc_apt_configure
from cloudinit.util import is_true
from tests.integration_tests.clouds import IntegrationCloud
from tests.integration_tests.instances import IntegrationInstance
from tests.integration_tests.integration_settings import (
    KEEP_INSTANCE,
    PLATFORM,
)
from tests.integration_tests.releases import (
    CURRENT_RELEASE,
    IS_UBUNTU,
    MANTIC,
)
from tests.integration_tests.util import (
    get_feature_flag_value,
    verify_clean_boot,
)

logger = logging.getLogger(__name__)

DEB822_SOURCES_FILE = "/etc/apt/sources.list.d/ubuntu.sources"
ORIG_SOURCES_FILE = "/etc/apt/sources.list"
GET_TEMPDIR = "python3 -c 'import tempfile;print(tempfile.mkdtemp());'"

USER_DATA = """\
#cloud-config
bootcmd:
    - rm -f /etc/apt/sources.list /etc/apt/sources.list.d/ubuntu.sources

apt:
  conf: |
    APT {
        Get {
            Assume-Yes "true";
            Fix-Broken "true";
        }
    }
  primary:
    - arches: [default]
      uri: http://badarchive.ubuntu.com/ubuntu
  security:
    - arches: [default]
      uri: http://badsecurity.ubuntu.com/ubuntu
  sources_list: |
    deb $MIRROR $RELEASE main restricted
    deb-src $MIRROR $RELEASE main restricted
    deb $PRIMARY $RELEASE universe restricted
    deb-src $PRIMARY $RELEASE universe restricted
    deb $SECURITY $RELEASE-security multiverse
    deb-src $SECURITY $RELEASE-security multiverse
  sources:
    test_keyserver:
      keyid: 1BC30F715A3B861247A81A5E55FE7C8C0165013E
      keyserver: keyserver.ubuntu.com
      # Hard-code noble as devel releases may not see new packages for some time
      source: "deb http://ppa.launchpad.net/curtin-dev/daily/ubuntu noble main"
    test_ppa:
      keyid: 3552C902B4DDF7BD38421821015D28D7441614D8
      keyserver: keyserver.ubuntu.com
      source: "ppa:simplestreams-dev/trunk"
    test_signed_by:
      keyid: A2EB2DEC0BD7519B7B38BE38376A290EC8068B11
      keyserver: keyserver.ubuntu.com
      source: "deb [signed-by=$KEY_FILE] http://ppa.launchpad.net/juju/stable/ubuntu $RELEASE main"
    test_bad_key:
      key: ""
      source: "deb $MIRROR $RELEASE main"
    test_key:
      source: "deb http://ppa.launchpad.net/cloud-init-dev/test-archive/ubuntu $RELEASE main"
      key: |
        -----BEGIN PGP PUBLIC KEY BLOCK-----
        Version: SKS 1.1.6
        Comment: Hostname: keyserver.ubuntu.com

        mQINBFbZRUIBEAC+A0PIKYBP9kLC4hQtRrffRS11uLo8/BdtmOdrlW0hpPHzCfKnjR3tvSEI
        lqPHG1QrrjAXKZDnZMRz+h/px7lUztvytGzHPSJd5ARUzAyjyRezUhoJ3VSCxrPqx62avuWf
        RfoJaIeHfDehL5/dTVkyiWxfVZ369ZX6JN2AgLsQTeybTQ75+2z0xPrrhnGmgh6g0qTYcAaq
        M5ONOGiqeSBX/Smjh6ALy5XkhUiFGLsI7Yluf6XSICY/x7gd6RAfgSIQrUTNMoS1sqhT4aot
        +xvOfQy8ySkfAK4NddXql6E/+ZqTmBY/Lr0YklFBy8jGT+UysfiIznPMIwbmgq5Li7BtDDtX
        b8Uyi4edPpjtextezfXYn4NVIpPL5dPZS/FXh4HpzyH0pYCfrH4QDGA7i52AGmhpiOFjJMo6
        N33sdjZHOH/2Vyp+QZaQnsdUAi1N4M6c33tQbpIScn1SY+El8z5JDA4PBzkw8HpLCi1gGoa6
        V4kfbWqXXbGAJFkLkP/vc4+pY9axOlmCkJg7xCPwhI75y1cONgovhz+BEXOzolh5KZuGbGbj
        xe0wva5DLBeIg7EQFf+99pOS7Syby3Xpm6ZbswEFV0cllK4jf/QMjtfInxobuMoI0GV0bE5l
        WlRtPCK5FnbHwxi0wPNzB/5fwzJ77r6HgPrR0OkT0lWmbUyoOQARAQABtC1MYXVuY2hwYWQg
        UFBBIGZvciBjbG91ZCBpbml0IGRldmVsb3BtZW50IHRlYW2JAjgEEwECACIFAlbZRUICGwMG
        CwkIBwMCBhUIAgkKCwQWAgMBAh4BAheAAAoJEAg9Bvvk0wTfHfcP/REK5N2s1JYc69qEa9ZN
        o6oi+A7l6AYw+ZY88O5TJe7F9otv5VXCIKSUT0Vsepjgf0mtXAgf/sb2lsJn/jp7tzgov3YH
        vSrkTkRydz8xcA87gwQKePuvTLxQpftF4flrBxgSueIn5O/tPrBOxLz7EVYBc78SKg9aj9L2
        yUp+YuNevlwfZCTYeBb9r3FHaab2HcgkwqYch66+nKYfwiLuQ9NzXXm0Wn0JcEQ6pWvJscbj
        C9BdawWovfvMK5/YLfI6Btm7F4mIpQBdhSOUp/YXKmdvHpmwxMCN2QhqYK49SM7qE9aUDbJL
        arppSEBtlCLWhRBZYLTUna+BkuQ1bHz4St++XTR49Qd7vDERALpApDjB2dxPfMiBzCMwQQyq
        uy13exU8o2ETLg+dZSLfDTzrBNsBFmXlw8WW17nTISYdKeGKL+QdlUjpzdwUMMzHhAO8SmMH
        zjeSlDSRMXBJFAFSbCl7EwmMKa3yVX0zInT91fNllZ3iatAmtVdqVH/BFQfTIMH2ET7A8WzJ
        ZzVSuMRhqoKdr5AMcHuJGPUoVkVJHQA+NNvEiXSysF3faL7jmKapmUwrhpYYX2H8pf+VMu2e
        cLflKTI28dl+ZQ4Pl/aVsxrti/pzhdYy05Sn5ddtySyIkvo8L1cU5MWpbvSlFPkTstBUDLBf
        pb0uBy+g0oxJQg15
        =uy53
        -----END PGP PUBLIC KEY BLOCK-----
    test_write:
      keyid: A2EB2DEC0BD7519B7B38BE38376A290EC8068B11
      keyserver: keyserver.ubuntu.com
      source: "deb [signed-by=$KEY_FILE] http://ppa.launchpad.net/juju/stable/ubuntu $RELEASE main"
      append: false
    test_write.list:
      keyid: A2EB2DEC0BD7519B7B38BE38376A290EC8068B11
      keyserver: keyserver.ubuntu.com
      source: "deb [signed-by=$KEY_FILE] http://ppa.launchpad.net/juju/devel/ubuntu $RELEASE main"
      append: false
    test_append:
      keyid: A2EB2DEC0BD7519B7B38BE38376A290EC8068B11
      keyserver: keyserver.ubuntu.com
      source: "deb [signed-by=$KEY_FILE] http://ppa.launchpad.net/juju/stable/ubuntu $RELEASE main"
    test_append.list:
      keyid: A2EB2DEC0BD7519B7B38BE38376A290EC8068B11
      keyserver: keyserver.ubuntu.com
      source: "deb [signed-by=$KEY_FILE] http://ppa.launchpad.net/juju/devel/ubuntu $RELEASE main"
apt_pipelining: os
"""  # noqa: E501

EXPECTED_REGEXES = [
    r"deb http://badarchive.ubuntu.com/ubuntu [a-z]+ main restricted",
    r"deb-src http://badarchive.ubuntu.com/ubuntu [a-z]+ main restricted",
    r"deb http://badarchive.ubuntu.com/ubuntu [a-z]+ universe restricted",
    r"deb-src http://badarchive.ubuntu.com/ubuntu [a-z]+ universe restricted",
    r"deb http://badsecurity.ubuntu.com/ubuntu [a-z]+-security multiverse",
    r"deb-src http://badsecurity.ubuntu.com/ubuntu [a-z]+-security multiverse",
]

TEST_KEYSERVER_KEY = "1BC3 0F71 5A3B 8612 47A8  1A5E 55FE 7C8C 0165 013E"
TEST_PPA_KEY = "3552 C902 B4DD F7BD 3842  1821 015D 28D7 4416 14D8"
TEST_KEY = "1FF0 D853 5EF7 E719 E5C8  1B9C 083D 06FB E4D3 04DF"
TEST_SIGNED_BY_KEY = "A2EB 2DEC 0BD7 519B 7B38  BE38 376A 290E C806 8B11"


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
@pytest.mark.user_data(USER_DATA)
class TestApt:
    def get_keys(self, class_client: IntegrationInstance):
        """Return all keys in /etc/apt/trusted.gpg.d/ and /etc/apt/trusted.gpg
        in human readable format. Mimics the output of apt-key finger
        """
        class_client.execute("mkdir /root/tmpdir && chmod &00 /root/tmpdir")
        GPG_LIST = [
            "gpg",
            "--no-options",
            "--with-fingerprint",
            "--homedir /root/tmpdir",
            "--no-default-keyring",
            "--list-keys",
            "--keyring",
        ]

        list_cmd = " ".join(GPG_LIST) + " "
        keys = class_client.execute(list_cmd + cc_apt_configure.APT_LOCAL_KEYS)
        files = class_client.execute(
            "ls " + cc_apt_configure.APT_TRUSTED_GPG_DIR
        ).stdout
        for file in files.split():
            path = cc_apt_configure.APT_TRUSTED_GPG_DIR + file
            keys += class_client.execute(list_cmd + path).stdout
        class_client.execute("gpgconf --homedir /root/tmpdir --kill all")
        return keys

    def test_sources_list(self, class_client: IntegrationInstance):
        """Integration test for the apt module's `sources_list` functionality.

        This test specifies a ``sources_list`` and then checks that (a) the
        expected number of sources.list entries is present, and (b) that each
        expected line appears in the file.

        Since sources_list is no deb822-compliant, ORIG_SOURCES_FILE will
        always be written regardless of feature.APT_DEB822_SOURCE_LIST_FILE
        """
        sources_list = class_client.read_from_file(ORIG_SOURCES_FILE)
        assert 6 == len(sources_list.rstrip().split("\n"))
        for expected_re in EXPECTED_REGEXES:
            assert re.search(expected_re, sources_list) is not None

    def test_apt_conf(self, class_client: IntegrationInstance):
        """Test the apt conf functionality."""
        apt_config = class_client.read_from_file(
            "/etc/apt/apt.conf.d/94cloud-init-config"
        )
        assert 'Assume-Yes "true";' in apt_config
        assert 'Fix-Broken "true";' in apt_config

    def test_ppa_source(self, class_client: IntegrationInstance):
        """Test the apt ppa functionality."""
        ppa_path = (
            "/etc/apt/sources.list.d/simplestreams-dev-ubuntu-trunk-{}".format(
                CURRENT_RELEASE.series
            )
        )
        if CURRENT_RELEASE < MANTIC:
            ppa_path += ".list"
        else:
            ppa_path += ".sources"
        ppa_path_contents = class_client.read_from_file(ppa_path)
        assert (
            "://ppa.launchpad.net/simplestreams-dev/trunk/ubuntu"
            in ppa_path_contents
            or "://ppa.launchpadcontent.net/simplestreams-dev/trunk/ubuntu"
            in ppa_path_contents
        )

        assert TEST_PPA_KEY in self.get_keys(class_client)

    def test_signed_by(self, class_client: IntegrationInstance):
        """Test the apt signed-by functionality."""
        source = (
            "deb [signed-by=/etc/apt/cloud-init.gpg.d/test_signed_by.gpg] "
            "http://ppa.launchpad.net/juju/stable/ubuntu"
            " {} main".format(CURRENT_RELEASE.series)
        )
        path_contents = class_client.read_from_file(
            "/etc/apt/sources.list.d/test_signed_by.list"
        )
        assert path_contents == source

        temp = class_client.execute(GET_TEMPDIR)
        key = class_client.execute(
            f"gpg --no-options --homedir {temp} --no-default-keyring "
            "--with-fingerprint --list-keys "
            "--keyring /etc/apt/cloud-init.gpg.d/test_signed_by.gpg"
        )

        assert TEST_SIGNED_BY_KEY in key

    def test_bad_key(self, class_client: IntegrationInstance):
        """Test the apt signed-by functionality."""
        with pytest.raises(OSError):
            class_client.read_from_file(
                "/etc/apt/trusted.list.d/test_bad_key.gpg"
            )

    def test_key(self, class_client: IntegrationInstance):
        """Test the apt key functionality."""
        test_archive_contents = class_client.read_from_file(
            "/etc/apt/sources.list.d/test_key.list"
        )

        assert (
            "http://ppa.launchpad.net/cloud-init-dev/test-archive/ubuntu"
            in test_archive_contents
        )
        assert TEST_KEY in self.get_keys(class_client)

    def test_keyserver(self, class_client: IntegrationInstance):
        """Test the apt keyserver functionality."""
        test_keyserver_contents = class_client.read_from_file(
            "/etc/apt/sources.list.d/test_keyserver.list"
        )

        assert (
            "http://ppa.launchpad.net/curtin-dev/daily/ubuntu"
            in test_keyserver_contents
        )

        assert TEST_KEYSERVER_KEY in self.get_keys(class_client)

    def test_os_pipelining(self, class_client: IntegrationInstance):
        """Test 'os' settings does not write apt config file."""
        conf_exists = class_client.execute(
            "test -f /etc/apt/apt.conf.d/90cloud-init-pipelining"
        ).ok
        assert conf_exists is False

    def test_sources_write(self, class_client: IntegrationInstance):
        """Test overwrite or append to sources file"""
        test_write_content = class_client.read_from_file(
            "/etc/apt/sources.list.d/test_write.list"
        )
        expected_contents = (
            "deb [signed-by=/etc/apt/cloud-init.gpg.d/test_write.gpg] "
            "http://ppa.launchpad.net/juju/devel/ubuntu "
            f"{CURRENT_RELEASE.series} main"
        )
        assert expected_contents.strip() == test_write_content.strip()

    def test_sources_append(self, class_client: IntegrationInstance):
        series = CURRENT_RELEASE.series
        test_append_content = class_client.read_from_file(
            "/etc/apt/sources.list.d/test_append.list"
        )

        expected_contents = (
            "deb [signed-by=/etc/apt/cloud-init.gpg.d/test_append.gpg] "
            f"http://ppa.launchpad.net/juju/stable/ubuntu {series} main\n"
            "deb [signed-by=/etc/apt/cloud-init.gpg.d/test_append.gpg] "
            f"http://ppa.launchpad.net/juju/devel/ubuntu {series} main"
        )
        assert expected_contents.strip() == test_append_content.strip()


_DEFAULT_DATA = """\
#cloud-config
apt:
  primary:
    - arches:
      - default
      {uri}
  security:
    - arches:
      - default
"""
DEFAULT_DATA = _DEFAULT_DATA.format(uri="")
# IBM sets sources_list in vendordata. Unset it
IBM_DEFAULT_DATA = DEFAULT_DATA + "\n  sources_list: ''\n"


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
@pytest.mark.user_data(DEFAULT_DATA)
class TestDefaults:
    @pytest.mark.skipif(
        PLATFORM != "openstack", reason="Test is Openstack specific"
    )
    def test_primary_on_openstack(self, class_client: IntegrationInstance):
        """Test apt default primary source on openstack.

        When no uri is provided.
        """
        zone = class_client.execute("cloud-init query v1.availability_zone")
        feature_deb822 = is_true(
            get_feature_flag_value(class_client, "APT_DEB822_SOURCE_LIST_FILE")
        )
        src_file = DEB822_SOURCES_FILE if feature_deb822 else ORIG_SOURCES_FILE
        sources_list = class_client.read_from_file(src_file)
        assert "{}.clouds.archive.ubuntu.com".format(zone) in sources_list

    @pytest.mark.skipif(
        PLATFORM == "ibm",
        reason="IBM has apt.sources_list in vendor-data overriding defaults",
    )
    def test_security(self, class_client: IntegrationInstance):
        """Test apt default security sources."""
        series = CURRENT_RELEASE.series
        feature_deb822 = is_true(
            get_feature_flag_value(class_client, "APT_DEB822_SOURCE_LIST_FILE")
        )
        if PLATFORM == "azure":
            sec_url = "http://azure.archive.ubuntu.com/ubuntu/"
        else:
            sec_url = "http://security.ubuntu.com/ubuntu"
        if feature_deb822:
            expected_cfg = dedent(
                f"""\
                Types: deb
                URIs: {sec_url}
                Suites: {series}-security
                """
            )
            sources_list = class_client.read_from_file(DEB822_SOURCES_FILE)
            assert expected_cfg in sources_list
        else:
            sources_list = class_client.read_from_file(ORIG_SOURCES_FILE)
            # 3 lines from main, universe, and multiverse
            sec_deb_line = f"deb {sec_url} {series}-security"
            sec_src_deb_line = sec_deb_line.replace("deb ", "# deb-src ")
            assert 3 == sources_list.count(sec_deb_line)
            assert 3 == sources_list.count(sec_src_deb_line)

    def test_no_duplicate_apt_sources(self, class_client: IntegrationInstance):
        r = class_client.execute("apt-get update", use_sudo=True)
        assert not re.match(
            r"^W: Target Packages .+ is configured multiple times in", r.stderr
        )

    def test_disabled_apt_sources(self, class_client: IntegrationInstance):
        feature_deb822 = is_true(
            get_feature_flag_value(class_client, "APT_DEB822_SOURCE_LIST_FILE")
        )
        if feature_deb822:

            assert (
                cc_apt_configure.UBUNTU_DEFAULT_APT_SOURCES_LIST.strip()
                == class_client.read_from_file(ORIG_SOURCES_FILE)
            )


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
@pytest.mark.skipif(
    PLATFORM != "ibm",
    reason="Overrides IBM-specific apt.sources_list in vendor-data",
)
@pytest.mark.user_data(IBM_DEFAULT_DATA)
def test_security_ibm(client: IntegrationInstance):
    """Test IBM APT default security sources."""
    series = CURRENT_RELEASE.series
    feature_deb822 = is_true(
        get_feature_flag_value(client, "APT_DEB822_SOURCE_LIST_FILE")
    )
    sec_url = "http://mirrors.adn.networklayer.com/ubuntu"
    if feature_deb822:
        expected_cfg = dedent(
            f"""\
            Types: deb
            URIs: {sec_url}
            Suites: {series}-security
            """
        )
        sources_list = client.read_from_file(DEB822_SOURCES_FILE)
        assert expected_cfg in sources_list
    else:
        sources_list = client.read_from_file(ORIG_SOURCES_FILE)
        # 3 lines from main, universe, and multiverse
        sec_deb_line = f"deb {sec_url} {series}-security"
        sec_src_deb_line = sec_deb_line.replace("deb ", "# deb-src ")
        assert 3 == sources_list.count(sec_deb_line)
        assert 3 == sources_list.count(sec_src_deb_line)


DEFAULT_DATA_WITH_URI = _DEFAULT_DATA.format(
    uri='uri: "http://something.random.invalid/ubuntu"'
)


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
def test_default_primary_with_uri(session_cloud: IntegrationCloud):
    """Test apt default primary sources."""
    userdata = DEFAULT_DATA_WITH_URI
    if PLATFORM == "ibm":
        # IBM provides apt.sources_list in vendordata. Unset it
        userdata += "\n  sources_list: ''\n"
    with session_cloud.launch(user_data=userdata) as client:
        feature_deb822 = is_true(
            get_feature_flag_value(client, "APT_DEB822_SOURCE_LIST_FILE")
        )
        src_file = DEB822_SOURCES_FILE if feature_deb822 else ORIG_SOURCES_FILE
        sources_list = client.read_from_file(src_file)
        assert "archive.ubuntu.com" not in sources_list
        assert "something.random.invalid" in sources_list


DISABLED_DATA = """\
#cloud-config
bootcmd: [mkdir -p /etc/apt/sources.new.d]
apt:
  conf: |
    Dir::Etc::sourceparts "sources.new.d";
  disable_suites:
  - $RELEASE
  - $RELEASE-updates
  - $RELEASE-backports
  - $RELEASE-security
apt_pipelining: false
"""


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
@pytest.mark.user_data(DISABLED_DATA)
class TestDisabled:
    def test_disable_suites(self, class_client: IntegrationInstance):
        """Test disabling of apt suites."""
        feature_deb822 = is_true(
            get_feature_flag_value(class_client, "APT_DEB822_SOURCE_LIST_FILE")
        )
        if feature_deb822:
            # DISABLED_DATA changes Dir:Etc::sourceparts to sources.new.d
            src_file = DEB822_SOURCES_FILE.replace("list", "new")
        else:
            src_file = ORIG_SOURCES_FILE
        sources_list = class_client.execute(
            f"cat {src_file} | grep -v '^#'"
        ).strip()
        assert "" == sources_list

    def test_disable_apt_pipelining(self, class_client: IntegrationInstance):
        """Test disabling of apt pipelining."""
        conf = class_client.read_from_file(
            "/etc/apt/apt.conf.d/90cloud-init-pipelining"
        )
        assert 'Acquire::http::Pipeline-Depth "0";' in conf


APT_PROXY_DATA = """\
#cloud-config
apt:
  proxy: "http://proxy.internal:3128"
  http_proxy: "http://squid.internal:3128"
  ftp_proxy: "ftp://squid.internal:3128"
  https_proxy: "https://squid.internal:3128"
"""


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
@pytest.mark.user_data(APT_PROXY_DATA)
def test_apt_proxy(client: IntegrationInstance):
    """Test the apt proxy data gets written correctly."""
    out = client.read_from_file("/etc/apt/apt.conf.d/90cloud-init-aptproxy")
    assert 'Acquire::http::Proxy "http://proxy.internal:3128";' in out
    assert 'Acquire::http::Proxy "http://squid.internal:3128";' in out
    assert 'Acquire::ftp::Proxy "ftp://squid.internal:3128";' in out
    assert 'Acquire::https::Proxy "https://squid.internal:3128";' in out


INSTALL_ANY_MISSING_RECOMMENDED_DEPENDENCIES = """\
#cloud-config
apt:
  sources:
    test_keyserver:
      keyid: 1BC30F715A3B861247A81A5E55FE7C8C0165013E
      keyserver: keyserver.ubuntu.com
      # Hard-code noble as devel releases may not see new packages for some time
      source: "deb http://ppa.launchpad.net/curtin-dev/daily/ubuntu noble main"
    test_ppa:
      keyid: 3552C902B4DDF7BD38421821015D28D7441614D8
      keyserver: keyserver.ubuntu.com
      source: "ppa:simplestreams-dev/trunk"
"""  # noqa: E501


RE_GPG_SW_PROPERTIES_INSTALLED = (
    r"install', '(gnupg', 'software-properties-common|"
    r"software-properties-common', 'gnupg)"
)

REMOVE_GPG_USERDATA = """
#cloud-config
runcmd:
  - DEBIAN_FRONTEND=noninteractive apt-get remove gpg -y
  - DEBIAN_FRONTEND=noninteractive apt-get remove software-properties-common -y
"""


def _do_oci_customization(cloud_config: str):
    """
    Add a snap disable command to the cloud-config for OCI.

    This is necessary to disable the oracle-cloud-agent snap on Oracle Cloud
    in order to prevent it from interfering with apt during tests.
    """
    addition = "  - snap disable oracle-cloud-agent"
    if PLATFORM == "oci":
        logger.info(
            "Running on Oracle Cloud, adding snap disable command to "
            "cloud-config to disable the oracle-cloud-agent snap."
        )
        return cloud_config.replace("runcmd:", f"runcmd:\n{addition}")
    return cloud_config


@pytest.mark.skipif(not IS_UBUNTU, reason="Apt usage")
def test_install_missing_deps(session_cloud: IntegrationCloud):
    """
    Test the installation of missing dependencies using apt on an Ubuntu
    system. This test is divided into two stages:
    Stage 1 (Remove existing packages):
    - Launch an instance with user-data that removes the 'gpg' and
      'software-properties-common' packages.
      - If on Oracle Cloud, add a command to the user-data to disable the
        oracle-cloud-agent snap to prevent it from interfering with apt.
    - Verify that the cloud-init log is clean and the boot process is clean.
    - Verify the packages are actually uninstalled using dpkg.
    - Create a snapshot of the instance after 'gpg' has been removed.
    - If KEEP_INSTANCE is False, destroy the instance after snapshotting.
    Stage 2 (re-install packages with user-data):
    - Launch a new instance from the snapshot created in Stage 1 with
      user-data that installs any missing recommended dependencies.
    - Verify that the cloud-init log is clean and the boot process is clean.
    - Check the cloud-init log to ensure that 'gpg' and
      'software-properties-common' are installed successfully.
    """
    # Two stage install: First stage:  remove gpg noninteractively from image
    instance1 = session_cloud.launch(
        user_data=_do_oci_customization(REMOVE_GPG_USERDATA)
    )

    # look for r"un  gpg" using regex ('un' means uninstalled)
    for package in ["gpg", "software-properties-common"]:
        dpkg_output = instance1.execute(
            "dpkg-query --show --showformat='${db:Status-Status}' " + package
        )
        assert dpkg_output.stdout in ("not-installed", "config-files"), (
            f"{package} package is still installed state {dpkg_output.stdout}."
            " It should have been removed by the user-data."
        )

    snapshot_id = instance1.snapshot()
    if not KEEP_INSTANCE:
        logger.info("Destroying instance1 after snapshotting.")
        instance1.destroy()
    else:
        logger.info(
            "Not destroying instance1 after snapshotting because KEEP_INSTANCE"
            " is True."
        )
    # Second stage: provide active apt user-data which will install missing gpg
    with session_cloud.launch(
        user_data=INSTALL_ANY_MISSING_RECOMMENDED_DEPENDENCIES,
        launch_kwargs={"image_id": snapshot_id},
    ) as minimal_client:
        log = minimal_client.read_from_file("/var/log/cloud-init.log")
        assert re.search(RE_GPG_SW_PROPERTIES_INSTALLED, log)
        gpg_installed = re.search(
            r"ii\s+gpg", minimal_client.execute("dpkg -l gpg").stdout
        )
        software_properties_common_installed = re.search(
            r"ii\s+software-properties-common",
            minimal_client.execute(
                "dpkg -l software-properties-common"
            ).stdout,
        )
        assert gpg_installed is not None, "gpg package is not installed."
        assert (
            software_properties_common_installed is not None
        ), "software-properties-common package is not installed."

        # It's a little weird that we're ignoring apt errors when testing apt,
        # but we've already verified that we install the missing dependencies.
        # To ensure `software-properties-common` is installed, we need to
        # specify a ppa in our user data, and `apt update` can fail if no ppa
        # has been uploaded for the release being tested. This isn't uncommon
        # for the devel release and newer releases in general.
        # Ignoring apt update errors seems preferable to playing whack-a-mole
        # with ppas that may or may not be available.
        verify_clean_boot(
            minimal_client,
            ignore_errors=["Failed to update package using apt"],
        )



# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init
import re
from collections import namedtuple
from typing import List
from unittest import mock

import pytest

from cloudinit import distros, helpers, subp, util
from cloudinit.config import cc_ca_certs
from cloudinit.config.schema import (
    SchemaValidationError,
    get_schema,
    validate_cloudconfig_schema,
)
from tests.unittests.helpers import SCHEMA_EMPTY_ERROR, skipUnlessJsonSchema
from tests.unittests.util import get_cloud

CAMocks = namedtuple("CAMocks", ["add", "update", "remove"])


class TestNoConfig:
    def test_no_config(self):
        """
        Test that nothing is done if no ca-certs configuration is provided.
        """
        name = "ca_certs"
        cloud_init = None
        args = []

        config = util.get_builtin_cfg()
        with mock.patch.object(util, "write_file") as util_mock:
            with mock.patch.object(
                cc_ca_certs, "update_ca_certs"
            ) as certs_mock:
                cc_ca_certs.handle(name, config, cloud_init, args)

                assert util_mock.call_count == 0
                assert certs_mock.call_count == 0


class TestConfig:
    name = "ca_certs"
    paths = None
    args: List = []

    def _fetch_distro(self, kind):
        cls = distros.fetch(kind)
        paths = helpers.Paths({})
        return cls(kind, {}, paths)

    @pytest.fixture
    def ca_mocks(self, mocker):
        """Fixture to provide mocked ca_cert functions"""
        mock_add = mocker.patch.object(cc_ca_certs, "add_ca_certs")
        mock_update = mocker.patch.object(cc_ca_certs, "update_ca_certs")
        mock_remove = mocker.patch.object(
            cc_ca_certs, "disable_default_ca_certs"
        )
        return CAMocks(mock_add, mock_update, mock_remove)

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_no_trusted_list(self, _, distro_name, ca_mocks):
        """
        Test that no certificates are written if the 'trusted' key is not
        present.
        """
        config = {"ca_certs": {}}
        cloud = get_cloud(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        assert ca_mocks.add.call_count == 0
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_empty_trusted_list(self, _, distro_name, ca_mocks):
        """Test that no certificate are written if 'trusted' list is empty."""
        config = {"ca_certs": {"trusted": []}}
        cloud = get_cloud(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        assert ca_mocks.add.call_count == 0
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_single_trusted(self, _, distro_name, ca_mocks):
        """Test that a single cert gets passed to add_ca_certs."""
        config = {"ca_certs": {"trusted": ["CERT1"]}}
        cloud = get_cloud(distro_name)
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        ca_mocks.add.assert_called_once_with(conf, ["CERT1"])
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_multiple_trusted(self, _, distro_name, ca_mocks):
        """Test that multiple certs get passed to add_ca_certs."""
        config = {"ca_certs": {"trusted": ["CERT1", "CERT2"]}}
        cloud = get_cloud(distro_name)
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        ca_mocks.add.assert_called_once_with(conf, ["CERT1", "CERT2"])
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_remove_default_ca_certs(self, _, distro_name, ca_mocks):
        """Test remove_defaults works as expected."""
        config = {"ca_certs": {"remove_defaults": True}}
        cloud = get_cloud(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        assert ca_mocks.add.call_count == 0
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 1

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_no_remove_defaults_if_false(self, _, distro_name, ca_mocks):
        """Test remove_defaults is not called when config value is False."""
        config = {"ca_certs": {"remove_defaults": False}}
        cloud = get_cloud(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        assert ca_mocks.add.call_count == 0
        assert ca_mocks.update.call_count == 1
        assert ca_mocks.remove.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    @mock.patch(
        "cloudinit.distros.networking.subp.subp",
        return_value=("", None),
    )
    def test_correct_order_for_remove_then_add(self, _, distro_name, ca_mocks):
        """
        Test remove_defaults is called before add.
        """
        config = {"ca_certs": {"remove_defaults": True, "trusted": ["CERT1"]}}
        cloud = get_cloud(distro_name)
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)
        cc_ca_certs.handle(self.name, config, cloud, self.args)

        assert ca_mocks.remove.call_count == 1
        ca_mocks.add.assert_called_once_with(conf, ["CERT1"])
        assert ca_mocks.update.call_count == 1


class TestAddCaCerts:
    @pytest.fixture
    def setup_test(self, tmp_path, mocker):
        """Fixture to set up test environment"""
        paths = helpers.Paths(
            {
                "cloud_dir": str(tmp_path),
            }
        )
        m_stat = mocker.patch("cloudinit.config.cc_ca_certs.os.stat")
        return paths, m_stat

    def _fetch_distro(self, kind):
        cls = distros.fetch(kind)
        paths = helpers.Paths({})
        return cls(kind, {}, paths)

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_no_certs_in_list(self, distro_name):
        """Test that no certificate are written if not provided."""
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)
        with mock.patch.object(util, "write_file") as mockobj:
            cc_ca_certs.add_ca_certs(conf, [])
        assert mockobj.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_single_cert(self, distro_name, setup_test):
        """Test adding a single certificate to the trusted CAs."""
        _, m_stat = setup_test
        cert = "CERT1\nLINE2\nLINE3"

        m_stat.return_value.st_size = 1
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)

        with mock.patch.object(util, "write_file") as mock_write:
            cc_ca_certs.add_ca_certs(conf, [cert])

            mock_write.assert_has_calls(
                [
                    mock.call(
                        conf["ca_cert_full_path"].format(cert_index=1),
                        cert,
                        mode=0o644,
                    )
                ]
            )

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_multiple_certs(self, distro_name, setup_test):
        """Test adding multiple certificates to the trusted CAs."""
        _, m_stat = setup_test
        certs = ["CERT1\nLINE2\nLINE3", "CERT2\nLINE2\nLINE3"]
        expected_cert_1_file = certs[0]
        expected_cert_2_file = certs[1]

        m_stat.return_value.st_size = 1
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)

        with mock.patch.object(util, "write_file") as mock_write:
            cc_ca_certs.add_ca_certs(conf, certs)

            mock_write.assert_has_calls(
                [
                    mock.call(
                        conf["ca_cert_full_path"].format(cert_index=1),
                        expected_cert_1_file,
                        mode=0o644,
                    ),
                    mock.call(
                        conf["ca_cert_full_path"].format(cert_index=2),
                        expected_cert_2_file,
                        mode=0o644,
                    ),
                ]
            )


class TestUpdateCaCerts:
    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_commands(self, distro_name):
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)
        with mock.patch.object(subp, "subp") as mockobj:
            cc_ca_certs.update_ca_certs(conf)
            mockobj.assert_called_once_with(
                conf["ca_cert_update_cmd"], capture=False
            )


class TestRemoveDefaultCaCerts:
    @pytest.fixture
    def setup_test(self, tmp_path, mocker):
        """Fixture to set up test environment"""
        paths = helpers.Paths(
            {
                "cloud_dir": str(tmp_path),
            }
        )
        m_stat = mocker.patch("cloudinit.config.cc_ca_certs.os.stat")
        return paths, m_stat

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_commands(self, distro_name, setup_test):
        _, m_stat = setup_test
        ca_certs_content = "# line1\nline2\nline3\n"
        expected = (
            "# line1\n# Modified by cloud-init to deselect certs due to"
            " user-data\n!line2\n!line3\n"
        )
        m_stat.return_value.st_size = 1
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)

        with mock.patch.object(util, "delete_dir_contents") as mock_delete:
            with mock.patch.object(
                util, "load_text_file", return_value=ca_certs_content
            ) as mock_load:
                with mock.patch.object(subp, "subp") as mock_subp:
                    with mock.patch.object(util, "write_file") as mock_write:
                        cc_ca_certs.disable_default_ca_certs(distro_name, conf)

                        if distro_name in ["rhel", "photon"]:
                            mock_delete.assert_has_calls(
                                [
                                    mock.call(conf["ca_cert_path"]),
                                    mock.call(conf["ca_cert_local_path"]),
                                ]
                            )
                            assert [] == mock_subp.call_args_list
                        elif distro_name in ["alpine", "debian", "ubuntu"]:
                            mock_load.assert_called_once_with(
                                conf["ca_cert_config"]
                            )
                            mock_write.assert_called_once_with(
                                conf["ca_cert_config"], expected, omode="wb"
                            )

                            if distro_name in ["debian", "ubuntu"]:
                                mock_subp.assert_called_once_with(
                                    ("debconf-set-selections", "-"),
                                    data=(
                                        "ca-certificates ca-certificates/"
                                        "trust_new_crts select no"
                                    ),
                                )
                            else:
                                assert mock_subp.call_count == 0

    @pytest.mark.parametrize("distro_name", cc_ca_certs.distros)
    def test_non_existent_cert_cfg(self, distro_name, setup_test):
        _, m_stat = setup_test
        m_stat.return_value.st_size = 0
        conf = cc_ca_certs._distro_ca_certs_configs(distro_name)

        with mock.patch.object(util, "delete_dir_contents"):
            with mock.patch.object(subp, "subp"):
                cc_ca_certs.disable_default_ca_certs(distro_name, conf)


@pytest.mark.usefixtures("clear_deprecation_log")
class TestCACertsSchema:
    """Directly test schema rather than through handle."""

    @pytest.mark.parametrize(
        "config, error_msg",
        (
            # Valid, yet deprecated schemas
            (
                {"ca-certs": {"remove-defaults": True}},
                re.escape(
                    "Cloud config schema deprecations: ca-certs:  "
                    "Deprecated in version 22.3. Use **ca_certs** instead.,"
                    " ca-certs.remove-defaults:  Deprecated in version 22.3"
                    ". Use **remove_defaults** instead."
                ),
            ),
            # Invalid schemas
            (
                {"ca_certs": 1},
                "ca_certs: 1 is not of type 'object'",
            ),
            (
                {"ca_certs": {}},
                re.escape("ca_certs: {} ") + SCHEMA_EMPTY_ERROR,
            ),
            (
                {"ca_certs": {"boguskey": 1}},
                re.escape(
                    "ca_certs: Additional properties are not allowed"
                    " ('boguskey' was unexpected)"
                ),
            ),
            (
                {"ca_certs": {"remove_defaults": 1}},
                "ca_certs.remove_defaults: 1 is not of type 'boolean'",
            ),
            (
                {"ca_certs": {"trusted": [1]}},
                "ca_certs.trusted.0: 1 is not of type 'string'",
            ),
            (
                {"ca_certs": {"trusted": []}},
                re.escape("ca_certs.trusted: [] ") + SCHEMA_EMPTY_ERROR,
            ),
        ),
    )
    @skipUnlessJsonSchema()
    def test_schema_validation(self, config, error_msg):
        """Assert expected schema validation and error messages."""
        # New-style schema $defs exist in config/cloud-init-schema*.json
        schema = get_schema()
        if error_msg is None:
            validate_cloudconfig_schema(config, schema, strict=True)
        else:
            with pytest.raises(SchemaValidationError, match=error_msg):
                validate_cloudconfig_schema(config, schema, strict=True)

    @mock.patch.object(cc_ca_certs, "update_ca_certs")
    def test_deprecate_key_warnings(self, update_ca_certs, caplog):
        """Assert warnings are logged for deprecated keys."""
        cloud = get_cloud("ubuntu")
        cc_ca_certs.handle(
            "IGNORE", {"ca-certs": {"remove-defaults": False}}, cloud, []
        )
        expected_warnings = [
            "Key 'ca-certs' is deprecated in",
            "Key 'remove-defaults' is deprecated in",
        ]
        for warning in expected_warnings:
            assert warning in caplog.text
            assert "deprecat" in caplog.text
        assert 1 == update_ca_certs.call_count

    @mock.patch.object(cc_ca_certs, "update_ca_certs")
    def test_duplicate_keys(self, update_ca_certs, caplog):
        """Assert warnings are logged for deprecated keys."""
        cloud = get_cloud("ubuntu")
        cc_ca_certs.handle(
            "IGNORE",
            {
                "ca-certs": {"remove-defaults": True},
                "ca_certs": {"remove_defaults": False},
            },
            cloud,
            [],
        )
        expected_warning = (
            "Found both ca-certs (deprecated) and ca_certs config keys."
            " Ignoring ca-certs."
        )
        assert expected_warning in caplog.text
        assert 1 == update_ca_certs.call_count



# This file is part of cloud-init. See LICENSE file for license information.
# pylint: disable=attribute-defined-outside-init
import json
from unittest import mock

import pytest
import responses

from cloudinit import distros, sources
from cloudinit import url_helper as uh
from cloudinit.atomic_helper import json_dumps
from cloudinit.sources import DataSourceCloudCIX as ds_mod
from cloudinit.sources import InvalidMetaDataException

METADATA = {
    "instance_id": "12_34",
    "network": {
        "version": 2,
        "ethernets": {
            "eth0": {
                "set-name": "eth0",
                "match": {"macaddress": "ab:cd:ef:00:01:02"},
                "addresses": [
                    "10.0.0.2/24",
                    "192.168.0.2/24",
                ],
                "nameservers": {
                    "addresses": ["10.0.0.1"],
                    "search": ["cloudcix.com"],
                },
                "routes": [{"to": "default", "via": "10.0.0.1"}],
            },
            "eth1": {
                "set-name": "eth1",
                "match": {"macaddress": "12:34:56:ab:cd:ef"},
                "addresses": [
                    "10.10.10.2/24",
                ],
                "nameservers": {
                    "addresses": ["10.0.0.1"],
                    "search": ["cloudcix.com"],
                },
            },
        },
    },
}

# Expected network config resulting from METADATA
NETWORK_CONFIG = {
    "version": 2,
    "ethernets": {
        "eth0": {
            "set-name": "eth0",
            "addresses": [
                "10.0.0.2/24",
                "192.168.0.2/24",
            ],
            "match": {"macaddress": "ab:cd:ef:00:01:02"},
            "nameservers": {
                "addresses": ["10.0.0.1"],
                "search": [
                    "cloudcix.com",
                ],
            },
            "routes": [{"to": "default", "via": "10.0.0.1"}],
        },
        "eth1": {
            "set-name": "eth1",
            "addresses": [
                "10.10.10.2/24",
            ],
            "match": {"macaddress": "12:34:56:ab:cd:ef"},
            "nameservers": {
                "addresses": ["10.0.0.1"],
                "search": [
                    "cloudcix.com",
                ],
            },
        },
    },
}

USERDATA = """#cloud-config
runcmd:
- [ echo Hello, World >> /etc/greeting ]
"""


class MockImds:
    @staticmethod
    def base_response(response):
        return 200, response.headers, "Metadata enabled"

    @staticmethod
    def metadata_response(response):
        return 200, response.headers, json.dumps(METADATA).encode()

    @staticmethod
    def userdata_response(response):
        return 200, response.headers, USERDATA.encode()


class TestDataSourceCloudCIX:
    """
    Test reading the meta-data
    """

    @pytest.fixture(autouse=True)
    def setup(self, mocker, tmpdir, paths):
        self.paths = paths
        self.datasource = self._get_ds()
        self.m_read_dmi_data = mocker.patch(
            "cloudinit.dmi.read_dmi_data",
            new_callable=mock.PropertyMock,
        )
        self.m_read_dmi_data.return_value = "CloudCIX"

        self._m_find_fallback_nic = mocker.patch(
            "cloudinit.net.find_fallback_nic",
            new_callable=mock.PropertyMock,
        )
        self._m_find_fallback_nic.return_value = "cixnic0"

    def _get_ds(self):
        distro_cls = distros.fetch("ubuntu")
        distro = distro_cls("ubuntu", cfg={}, paths=self.paths)
        return ds_mod.DataSourceCloudCIX(
            sys_cfg={
                "datasource": {
                    "CloudCIX": {
                        "retries": 1,
                        "timeout": 0,
                        "wait": 0,
                    },
                }
            },
            distro=distro,
            paths=self.paths,
        )

    @responses.activate
    def test_identifying_cloudcix(self):
        assert self.datasource.ds_detect()
        assert ds_mod.is_platform_viable()

        self.m_read_dmi_data.return_value = "OnCloud9"
        assert not self.datasource.ds_detect()
        assert not ds_mod.is_platform_viable()

    def test_setting_config_options(self):
        cix_options = {
            "timeout": 1234,
            "retries": 5678,
            "sec_between_retries": 9012,
        }
        sys_cfg = {
            "datasource": {
                "CloudCIX": cix_options,
            }
        }

        # Instantiate a new datasource
        distro_cls = distros.fetch("ubuntu")
        distro = distro_cls("ubuntu", cfg={}, paths=self.paths)
        new_ds = ds_mod.DataSourceCloudCIX(
            sys_cfg=sys_cfg, distro=distro, paths=self.paths
        )
        assert (
            new_ds.get_url_params().timeout_seconds == cix_options["timeout"]
        )
        assert new_ds.get_url_params().num_retries == cix_options["retries"]
        assert (
            new_ds.get_url_params().sec_between_retries
            == cix_options["sec_between_retries"]
        )

    @responses.activate
    def test_determine_md_url(self):
        base_url = ds_mod.METADATA_URLS[0]
        version = ds_mod.METADATA_VERSION
        responses.reset()
        responses.add_callback(
            responses.GET,
            base_url,
            callback=MockImds.base_response,
        )
        responses.add_callback(
            responses.GET,
            uh.combine_url(base_url, f"v{version}", "metadata"),
            callback=MockImds.metadata_response,
        )

        md_url = self._get_ds().determine_md_url()

        expected_url = uh.combine_url(
            ds_mod.METADATA_URLS[0],
            f"v{version}",
        )
        assert md_url == expected_url

    @responses.activate
    def test_reading_metadata_on_cloudcix(self):
        base_url = ds_mod.METADATA_URLS[0]
        version = ds_mod.METADATA_VERSION
        # Set up mock endpoints
        responses.reset()
        responses.add_callback(
            responses.GET,
            base_url,
            callback=MockImds.base_response,
        )
        responses.add_callback(
            responses.GET,
            uh.combine_url(base_url, f"v{version}", "metadata"),
            callback=MockImds.metadata_response,
        )
        responses.add_callback(
            responses.GET,
            uh.combine_url(base_url, f"v{version}", "userdata"),
            callback=MockImds.userdata_response,
        )

        assert self.datasource.get_data()
        assert self.datasource.metadata == METADATA
        assert self.datasource.userdata_raw == USERDATA
        assert json_dumps(self.datasource.network_config) == json_dumps(
            NETWORK_CONFIG
        )

    @responses.activate
    def test_failing_imds_endpoints(self, mocker):
        sleep = mocker.patch("time.sleep")
        base_url = ds_mod.METADATA_URLS[0]
        # Make request before imds is set up
        with pytest.raises(
            sources.InvalidMetaDataException,
            match="Could not determine metadata URL",
        ):
            self.datasource.crawl_metadata_service()

        # Make imds respond to healthcheck but fail v1/metadata
        responses.add_callback(
            responses.GET,
            base_url,
            callback=MockImds.base_response,
        )
        version = ds_mod.METADATA_VERSION
        with pytest.raises(
            sources.InvalidMetaDataException,
            match="Could not determine metadata URL",
        ):
            self.datasource.crawl_metadata_service()

        # No sleep/retries when md_url returns 404. No viable IMDS found.
        assert 0 == sleep.call_count

        # Make imds serve metadata but ConnectionError on userdata
        responses.add_callback(
            responses.GET,
            uh.combine_url(base_url, f"v{version}", "metadata"),
            callback=MockImds.metadata_response,
        )
        pytest.raises(
            sources.InvalidMetaDataException,
            self.datasource.crawl_metadata_service,
        )
        # Sleep called number of default datasource configured "retries"
        assert [mock.call(5)] == sleep.call_args_list

        # Make IMDS serve userdata
        responses.add_callback(
            responses.GET,
            uh.combine_url(base_url, f"v{version}", "userdata"),
            callback=MockImds.userdata_response,
        )

        data = self.datasource.crawl_metadata_service()
        assert data == {
            "meta-data": METADATA,
            "user-data": USERDATA.encode("utf-8"),
        }

    @responses.activate
    def test_read_malformed_metadata(self):
        def bad_response(response):
            return 200, response.headers, json.dumps(METADATA)[:-2]

        version = ds_mod.METADATA_VERSION
        base_url = ds_mod.METADATA_URLS[0]
        versioned_url = uh.combine_url(base_url, f"v{version}")

        # Malformed metadata
        responses.add_callback(
            responses.GET,
            uh.combine_url(versioned_url, "metadata"),
            callback=bad_response,
        )
        responses.add_callback(
            responses.GET,
            uh.combine_url(versioned_url, "userdata"),
            callback=MockImds.userdata_response,
        )

        with pytest.raises(
            InvalidMetaDataException,
            match="Invalid JSON at http://169.254.169.254/v1/metadata",
        ):
            ds_mod.read_metadata(
                versioned_url,
                self.datasource.get_url_params(),
            )

    @responses.activate
    def test_bad_response_code(self, mocker):
        def bad_response(response):
            return 404, response.headers, ""

        version = ds_mod.METADATA_VERSION
        base_url = ds_mod.METADATA_URLS[0]
        versioned_url = uh.combine_url(base_url, f"v{version}")

        responses.add_callback(
            responses.GET,
            uh.combine_url(versioned_url, "metadata"),
            callback=bad_response,
        )
        sleep = mocker.patch("time.sleep")
        with pytest.raises(
            InvalidMetaDataException,
            match=f"Failed to fetch IMDS metadata: {versioned_url}",
        ):
            ds_mod.read_metadata(
                versioned_url, self.datasource.get_url_params()
            )
        assert [mock.call(5)] == sleep.call_args_list



# (c) Copyright IBM Corp. 2020 All Rights Reserved
#
# Author: Aman Kumar Sinha <amansi26@in.ibm.com>
#
# This file is part of cloud-init. See LICENSE file for license information.
"""Reset RMC: Reset rsct node id

Reset RMC module is IBM PowerVM Hypervisor specific

Reliable Scalable Cluster Technology (RSCT) is a set of software components,
that  together provide a comprehensive clustering environment (RAS features)
for IBM PowerVM based virtual machines. RSCT includes the Resource monitoring
and control (RMC) subsystem. RMC is a generalized framework used for managing,
monitoring, and manipulating resources. RMC runs as a daemon process on
individual machines and needs creation of unique node id and restarts
during VM boot.
More details refer
https://www.ibm.com/support/knowledgecenter/en/SGVKBA_3.2/admin/bl503_ovrv.htm

This module handles
- creation of the unique RSCT node id to every instance/virtual machine
  and ensure once set, it isn't changed subsequently by cloud-init.
  In order to do so, it restarts RSCT service.

Prerequisite of using this module is to install RSCT packages.
"""
import logging
import os

from cloudinit import subp, util
from cloudinit.cloud import Cloud
from cloudinit.config import Config
from cloudinit.config.schema import MetaSchema
from cloudinit.distros import ALL_DISTROS
from cloudinit.settings import PER_INSTANCE

meta: MetaSchema = {
    "id": "cc_reset_rmc",
    "distros": [ALL_DISTROS],
    "frequency": PER_INSTANCE,
    "activate_by_schema_keys": [],
}

# This module is undocumented in our schema docs
__doc__ = ""

# RMCCTRL is expected to be in system PATH (/opt/rsct/bin)
# The symlink for RMCCTRL and RECFGCT are
# /usr/sbin/rsct/bin/rmcctrl and
# /usr/sbin/rsct/install/bin/recfgct respectively.
RSCT_PATH = "/opt/rsct/install/bin"
RMCCTRL = "rmcctrl"
RECFGCT = "recfgct"

LOG = logging.getLogger(__name__)

NODE_ID_FILE = "/etc/ct_node_id"


def handle(name: str, cfg: Config, cloud: Cloud, args: list) -> None:
    # Ensuring node id has to be generated only once during first boot
    if cloud.datasource.platform_type == "none":
        LOG.debug("Skipping creation of new ct_node_id node")
        return

    if not os.path.isdir(RSCT_PATH):
        LOG.debug("module disabled, RSCT_PATH not present")
        return

    orig_path = os.environ.get("PATH")
    try:
        add_path(orig_path)
        reset_rmc()
    finally:
        if orig_path:
            os.environ["PATH"] = orig_path
        else:
            del os.environ["PATH"]


def reconfigure_rsct_subsystems():
    # Reconfigure the RSCT subsystems, which includes removing all RSCT data
    # under the /var/ct directory, generating a new node ID, and making it
    # appear as if the RSCT components were just installed
    try:
        out = subp.subp([RECFGCT])[0]
        LOG.debug(out.strip())
        return out
    except subp.ProcessExecutionError:
        util.logexc(LOG, "Failed to reconfigure the RSCT subsystems.")
        raise


def get_node_id():
    try:
        fp = util.load_text_file(NODE_ID_FILE)
        node_id = fp.split("\n")[0]
        return node_id
    except Exception:
        util.logexc(LOG, "Failed to get node ID from file %s." % NODE_ID_FILE)
        raise


def add_path(orig_path):
    # Adding the RSCT_PATH to env standard path
    # So that cloud init automatically find and
    # run RECFGCT to create new node_id.
    suff = ":" + orig_path if orig_path else ""
    os.environ["PATH"] = RSCT_PATH + suff
    return os.environ["PATH"]


def rmcctrl():
    # Stop the RMC subsystem and all resource managers so that we can make
    # some changes to it
    try:
        return subp.subp([RMCCTRL, "-z"])
    except Exception:
        util.logexc(LOG, "Failed to stop the RMC subsystem.")
        raise


def reset_rmc():
    LOG.debug("Attempting to reset RMC.")

    node_id_before = get_node_id()
    LOG.debug("Node ID at beginning of module: %s", node_id_before)

    # Stop the RMC subsystem and all resource managers so that we can make
    # some changes to it
    rmcctrl()
    reconfigure_rsct_subsystems()

    node_id_after = get_node_id()
    LOG.debug("Node ID at end of module: %s", node_id_after)

    # Check if new node ID is generated or not
    # by comparing old and new node ID
    if node_id_after == node_id_before:
        msg = "New node ID did not get generated."
        LOG.error(msg)
        raise RuntimeError(msg)



# Copyright (c) 2021-2025 Broadcom. All Rights Reserved.
#
# Authors: Andrew Kutz <andrew.kutz@broadcom.com>
#          Pengpeng Sun <pengpeng.sun@broadcom.com>
#
# This file is part of cloud-init. See LICENSE file for license information.

import base64
import gzip
import os
from contextlib import ExitStack
from logging import DEBUG
from textwrap import dedent

import pytest

from cloudinit import dmi, safeyaml, settings, util
from cloudinit.event import EventScope
from cloudinit.sources import DataSourceVMware
from cloudinit.sources.helpers.vmware.imc import guestcust_util
from cloudinit.subp import ProcessExecutionError
from tests.unittests.helpers import mock, populate_dir, wrap_and_call

MPATH = "cloudinit.sources.DataSourceVMware."
PRODUCT_NAME_FILE_PATH = "/sys/class/dmi/id/product_name"
PRODUCT_NAME = "VMware7,1"
PRODUCT_UUID = "82343CED-E4C7-423B-8F6B-0D34D19067AB"
REROOT_FILES = {
    DataSourceVMware.PRODUCT_UUID_FILE_PATH: PRODUCT_UUID,
    PRODUCT_NAME_FILE_PATH: PRODUCT_NAME,
}

VMW_MULTIPLE_KEYS = [
    "ssh-rsa AAAAB3NzaC1yc2EAAAA... test1@vmw.com",
    "ssh-rsa AAAAB3NzaC1yc2EAAAA... test2@vmw.com",
]
VMW_SINGLE_KEY = "ssh-rsa AAAAB3NzaC1yc2EAAAA... test@vmw.com"

VMW_METADATA_YAML = """\
instance-id: cloud-vm
local-hostname: cloud-vm
network:
  version: 2
  ethernets:
    nics:
      match:
        name: ens*
      dhcp4: yes
"""

VMW_USERDATA_YAML = """\
## template: jinja
#cloud-config
users:
- default
"""

VMW_VENDORDATA_YAML = """\
## template: jinja
#cloud-config
runcmd:
- echo "Hello, world."
"""

VMW_IPV4_ROUTEINFO = {
    "destination": "0.0.0.0",
    "flags": "G",
    "gateway": "10.85.130.1",
    "genmask": "0.0.0.0",
    "iface": "eth0",
    "metric": "50",
}
VMW_IPV4_NETDEV_ADDR = {
    "bcast": "10.85.130.255",
    "ip": "10.85.130.116",
    "mask": "255.255.255.0",
    "scope": "global",
}
VMW_IPV4_NETIFACES_ADDR = {
    "broadcast": "10.85.130.255",
    "netmask": "255.255.255.0",
    "addr": "10.85.130.116",
}
VMW_IPV6_ROUTEINFO = {
    "destination": "::/0",
    "flags": "UG",
    "gateway": "2001:67c:1562:8007::1",
    "iface": "eth0",
    "metric": "50",
}
VMW_IPV6_NETDEV_ADDR = {
    "ip": "fd42:baa2:3dd:17a:216:3eff:fe16:db54/64",
    "scope6": "global",
}
VMW_IPV6_NETIFACES_ADDR = {
    "netmask": "ffff:ffff:ffff:ffff::/64",
    "addr": "fd42:baa2:3dd:17a:216:3eff:fe16:db54",
}
VMW_IPV6_NETDEV_PEER_ADDR = {
    "ip": "fd42:baa2:3dd:17a:216:3eff:fe16:db54",
    "scope6": "global",
}
VMW_IPV6_NETIFACES_PEER_ADDR = {
    "netmask": "ffff:ffff:ffff:ffff:ffff:ffff:ffff:ffff/128",
    "addr": "fd42:baa2:3dd:17a:216:3eff:fe16:db54",
}

# Please note this should be a constant, but uses formatting to avoid
# the line-length warning from the linter.
VMW_EXPECTED_EXTRA_HOTPLUG_UDEV_RULES = """
ENV{ID_NET_DRIVER}=="e1000|e1000e|vlance|vmxnet2|vmxnet3|vrdma", GOTO="cloudinit_hook"
GOTO="cloudinit_end"
"""  # noqa: E501


VMW_METADATA_YAML_WITH_NET_DRIVERS = """\
instance-id: cloud-vm
local-hostname: cloud-vm
network-drivers:
- vmxnet2
- vmxnet3
"""

VMW_EXPECTED_EXTRA_HOTPLUG_UDEV_RULES_VMXNET = """
ENV{ID_NET_DRIVER}=="vmxnet2|vmxnet3", GOTO="cloudinit_hook"
GOTO="cloudinit_end"
"""


def generate_test_netdev_data(ipv4=None, ipv6=None):
    ipv4 = ipv4 or []
    ipv6 = ipv6 or []
    return {
        "eth0": {
            "hwaddr": "00:16:3e:16:db:54",
            "ipv4": ipv4,
            "ipv6": ipv6,
            "up": True,
        },
    }


@pytest.fixture(autouse=True)
def common_patches():
    mocks = [
        mock.patch("cloudinit.util.platform.platform", return_value="Linux"),
        mock.patch.multiple(
            "cloudinit.dmi",
            is_container=mock.Mock(return_value=False),
            is_FreeBSD=mock.Mock(return_value=False),
        ),
        mock.patch(
            "cloudinit.netinfo.netdev_info",
            return_value={},
        ),
        mock.patch(
            "cloudinit.sources.DataSourceVMware.getfqdn",
            return_value="host.cloudinit.test",
        ),
    ]
    with ExitStack() as stack:
        for some_mock in mocks:
            stack.enter_context(some_mock)
        yield


class TestDataSourceVMware:
    """
    Test common functionality that is not transport specific.
    """

    def test_no_data_access_method(self, DS):
        ds = DS(settings.CFG_BUILTIN)
        with mock.patch(
            "cloudinit.sources.DataSourceVMware.is_vmware_platform",
            return_value=False,
        ):
            ret = ds.get_data()
        assert not ret

    def test_convert_to_netifaces_ipv4_format(self):
        netifaces_format = DataSourceVMware.convert_to_netifaces_ipv4_format(
            VMW_IPV4_NETDEV_ADDR
        )
        assert netifaces_format == VMW_IPV4_NETIFACES_ADDR

    def test_convert_to_netifaces_ipv6_format(self):
        netifaces_format = DataSourceVMware.convert_to_netifaces_ipv6_format(
            VMW_IPV6_NETDEV_ADDR
        )
        assert netifaces_format == VMW_IPV6_NETIFACES_ADDR
        netifaces_format = DataSourceVMware.convert_to_netifaces_ipv6_format(
            VMW_IPV6_NETDEV_PEER_ADDR
        )
        assert netifaces_format == VMW_IPV6_NETIFACES_PEER_ADDR

    @mock.patch("cloudinit.sources.DataSourceVMware.get_default_ip_addrs")
    def test_get_host_info_ipv4(self, m_fn_ipaddr):
        m_fn_ipaddr.return_value = ("10.10.10.1", None)
        host_info = DataSourceVMware.get_host_info()
        assert host_info
        assert host_info["hostname"]
        assert host_info["hostname"] == "host.cloudinit.test"
        assert host_info["local-hostname"]
        assert host_info["local_hostname"]
        assert host_info[DataSourceVMware.LOCAL_IPV4]
        assert host_info[DataSourceVMware.LOCAL_IPV4] == "10.10.10.1"
        assert not host_info.get(DataSourceVMware.LOCAL_IPV6)

    @mock.patch("cloudinit.sources.DataSourceVMware.get_default_ip_addrs")
    def test_get_host_info_ipv6(self, m_fn_ipaddr):
        m_fn_ipaddr.return_value = (None, "2001:db8::::::8888")
        host_info = DataSourceVMware.get_host_info()
        assert host_info
        assert host_info["hostname"]
        assert host_info["hostname"] == "host.cloudinit.test"
        assert host_info["local-hostname"]
        assert host_info["local_hostname"]
        assert host_info[DataSourceVMware.LOCAL_IPV6]
        assert host_info[DataSourceVMware.LOCAL_IPV6] == "2001:db8::::::8888"
        assert not host_info.get(DataSourceVMware.LOCAL_IPV4)

    @mock.patch("cloudinit.sources.DataSourceVMware.get_default_ip_addrs")
    def test_get_host_info_dual(self, m_fn_ipaddr):
        m_fn_ipaddr.return_value = ("10.10.10.1", "2001:db8::::::8888")
        host_info = DataSourceVMware.get_host_info()
        assert host_info
        assert host_info["hostname"]
        assert host_info["hostname"] == "host.cloudinit.test"
        assert host_info["local-hostname"]
        assert host_info["local_hostname"]
        assert host_info[DataSourceVMware.LOCAL_IPV4]
        assert host_info[DataSourceVMware.LOCAL_IPV4] == "10.10.10.1"
        assert host_info[DataSourceVMware.LOCAL_IPV6]
        assert host_info[DataSourceVMware.LOCAL_IPV6] == "2001:db8::::::8888"

    # TODO migrate this entire test suite to pytest then parameterize
    @mock.patch("cloudinit.netinfo.route_info")
    @mock.patch("cloudinit.netinfo.netdev_info")
    def test_get_default_ip_addrs_ipv4only(
        self,
        m_netdev_info,
        m_route_info,
    ):
        """Test get_default_ip_addrs use cases"""
        m_route_info.return_value = {
            "ipv4": [VMW_IPV4_ROUTEINFO],
            "ipv6": [],
        }
        m_netdev_info.return_value = generate_test_netdev_data(
            ipv4=[VMW_IPV4_NETDEV_ADDR]
        )
        ipv4, ipv6 = DataSourceVMware.get_default_ip_addrs()
        assert ipv4 == "10.85.130.116"
        assert ipv6 is None

    @mock.patch("cloudinit.netinfo.route_info")
    @mock.patch("cloudinit.netinfo.netdev_info")
    def test_get_default_ip_addrs_ipv6only(
        self,
        m_netdev_info,
        m_route_info,
    ):
        m_route_info.return_value = {
            "ipv4": [],
            "ipv6": [VMW_IPV6_ROUTEINFO],
        }
        m_netdev_info.return_value = generate_test_netdev_data(
            ipv6=[VMW_IPV6_NETDEV_ADDR]
        )
        ipv4, ipv6 = DataSourceVMware.get_default_ip_addrs()
        assert ipv4 is None
        assert ipv6 == "fd42:baa2:3dd:17a:216:3eff:fe16:db54/64"

    @mock.patch("cloudinit.netinfo.route_info")
    @mock.patch("cloudinit.netinfo.netdev_info")
    def test_get_default_ip_addrs_dualstack(
        self,
        m_netdev_info,
        m_route_info,
    ):
        m_route_info.return_value = {
            "ipv4": [VMW_IPV4_ROUTEINFO],
            "ipv6": [VMW_IPV6_ROUTEINFO],
        }
        m_netdev_info.return_value = generate_test_netdev_data(
            ipv4=[VMW_IPV4_NETDEV_ADDR],
            ipv6=[VMW_IPV6_NETDEV_ADDR],
        )
        ipv4, ipv6 = DataSourceVMware.get_default_ip_addrs()
        assert ipv4 == "10.85.130.116"
        assert ipv6 == "fd42:baa2:3dd:17a:216:3eff:fe16:db54/64"

    @mock.patch("cloudinit.netinfo.route_info")
    @mock.patch("cloudinit.netinfo.netdev_info")
    def test_get_default_ip_addrs_multiaddr(
        self,
        m_netdev_info,
        m_route_info,
    ):
        m_route_info.return_value = {
            "ipv4": [VMW_IPV4_ROUTEINFO],
            "ipv6": [],
        }
        m_netdev_info.return_value = generate_test_netdev_data(
            ipv4=[
                VMW_IPV4_NETDEV_ADDR,
                {
                    "bcast": "10.85.131.255",
                    "ip": "10.85.131.117",
                    "mask": "255.255.255.0",
                    "scope": "global",
                },
            ],
            ipv6=[
                VMW_IPV6_NETDEV_ADDR,
                {
                    "ip": "fe80::216:3eff:fe16:db54/64",
                    "scope6": "link",
                },
            ],
        )
        ipv4, ipv6 = DataSourceVMware.get_default_ip_addrs()
        assert ipv4 is None
        assert ipv6 is None

    @mock.patch("cloudinit.netinfo.route_info")
    @mock.patch("cloudinit.netinfo.netdev_info")
    def test_get_default_ip_addrs_nodefault(
        self,
        m_netdev_info,
        m_route_info,
    ):
        m_route_info.return_value = {
            "ipv4": [
                {
                    "destination": "185.125.188.0",
                    "flags": "G",
                    "gateway": "10.85.130.1",
                    "genmask": "0.0.0.255",
                    "iface": "eth0",
                    "metric": "50",
                },
            ],
            "ipv6": [],
        }
        m_netdev_info.return_value = generate_test_netdev_data(
            ipv4=[VMW_IPV4_NETDEV_ADDR],
            ipv6=[VMW_IPV6_NETDEV_ADDR],
        )
        ipv4, ipv6 = DataSourceVMware.get_default_ip_addrs()
        assert ipv4 is None
        assert ipv6 is None

    @mock.patch("cloudinit.sources.DataSourceVMware.get_host_info")
    def test_wait_on_network(self, m_fn, caplog):
        metadata = {
            DataSourceVMware.WAIT_ON_NETWORK: {
                DataSourceVMware.WAIT_ON_NETWORK_IPV4: True,
                DataSourceVMware.WAIT_ON_NETWORK_IPV6: False,
            },
        }
        m_fn.side_effect = [
            {
                "hostname": "host.cloudinit.test",
                "local-hostname": "host.cloudinit.test",
                "local_hostname": "host.cloudinit.test",
                "network": {
                    "interfaces": {
                        "by-ipv4": {},
                        "by-ipv6": {},
                        "by-mac": {
                            "aa:bb:cc:dd:ee:ff": {"ipv4": [], "ipv6": []}
                        },
                    },
                },
            },
            {
                "hostname": "host.cloudinit.test",
                "local-hostname": "host.cloudinit.test",
                "local-ipv4": "10.10.10.1",
                "local_hostname": "host.cloudinit.test",
                "network": {
                    "interfaces": {
                        "by-ipv4": {
                            "10.10.10.1": {
                                "mac": "aa:bb:cc:dd:ee:ff",
                                "netmask": "255.255.255.0",
                            }
                        },
                        "by-mac": {
                            "aa:bb:cc:dd:ee:ff": {
                                "ipv4": [
                                    {
                                        "addr": "10.10.10.1",
                                        "broadcast": "10.10.10.255",
                                        "netmask": "255.255.255.0",
                                    }
                                ],
                                "ipv6": [],
                            }
                        },
                    },
                },
            },
        ]

        host_info = DataSourceVMware.wait_on_network(metadata)

        expected_logs = [
            (
                "cloudinit.sources.DataSourceVMware",
                DEBUG,
                (
                    "waiting on network: wait4=True, "
                    "ready4=False, wait6=False, ready6=False"
                ),
            ),
            (
                "cloudinit.sources.DataSourceVMware",
                DEBUG,
                "waiting on network complete",
            ),
        ]
        for log in expected_logs:
            assert log in caplog.record_tuples

        assert host_info
        assert host_info["hostname"]
        assert host_info["hostname"] == "host.cloudinit.test"
        assert host_info["local-hostname"]
        assert host_info["local_hostname"]
        assert host_info[DataSourceVMware.LOCAL_IPV4]
        assert host_info[DataSourceVMware.LOCAL_IPV4] == "10.10.10.1"

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_set_value")
    def test_advertise_update_events(self, m_set_fn):
        (
            supported_events,
            enabled_events,
        ) = DataSourceVMware.advertise_update_events(
            DataSourceVMware.SUPPORTED_UPDATE_EVENTS,
            DataSourceVMware.DEFAULT_UPDATE_EVENTS,
            "rpctool",
            len,
        )
        assert 2 == m_set_fn.call_count
        assert "network=boot;boot-new-instance;hotplug" == supported_events
        assert "network=boot-new-instance;hotplug" == enabled_events

    def test_extra_hotplug_udev_rules(self, DS):
        ds = DS(settings.CFG_BUILTIN)
        assert (
            VMW_EXPECTED_EXTRA_HOTPLUG_UDEV_RULES
            == ds.extra_hotplug_udev_rules
        )


class TestDataSourceVMwareEnvVars:
    """
    Test the envvar transport.
    """

    @pytest.fixture(autouse=True)
    def env_and_files(self, fake_filesystem, monkeypatch, tmpdir):
        monkeypatch.setenv(DataSourceVMware.VMX_GUESTINFO, "1")
        populate_dir(
            str(tmpdir),
            {DataSourceVMware.PRODUCT_UUID_FILE_PATH: PRODUCT_UUID},
        )

    def assert_get_data_ok(self, DS, m_fn, m_fn_call_count=6):
        ds = DS(settings.CFG_BUILTIN)
        ret = ds.get_data()
        assert ret
        assert m_fn_call_count == m_fn.call_count
        assert (
            ds.data_access_method == DataSourceVMware.DATA_ACCESS_METHOD_ENVVAR
        )
        return ds

    def assert_metadata(self, DS, metadata, m_fn, m_fn_call_count=6):
        ds = self.assert_get_data_ok(DS, m_fn, m_fn_call_count)
        assert_metadata(ds, metadata)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_subplatform(self, m_fn, DS):
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        ds = self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)
        assert ds.subplatform == "%s (%s)" % (
            DataSourceVMware.DATA_ACCESS_METHOD_ENVVAR,
            DataSourceVMware.get_guestinfo_envvar_key_name("metadata"),
        )

        # Test to ensure that network is configured from metadata on each boot.
        assert (
            DataSourceVMware.DEFAULT_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.default_update_events[EventScope.NETWORK]
        )
        assert (
            DataSourceVMware.SUPPORTED_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.supported_update_events[EventScope.NETWORK]
        )

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_metadata_only(self, m_fn, DS):
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_userdata_only(self, m_fn, DS):
        m_fn.side_effect = ["", VMW_USERDATA_YAML, "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_vendordata_only(self, m_fn, DS):
        m_fn.side_effect = ["", "", VMW_VENDORDATA_YAML, ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_metadata_base64(self, m_fn, DS):
        data = base64.b64encode(VMW_METADATA_YAML.encode("utf-8"))
        m_fn.side_effect = [data, "base64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_metadata_b64(self, m_fn, DS):
        data = base64.b64encode(VMW_METADATA_YAML.encode("utf-8"))
        m_fn.side_effect = [data, "b64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_metadata_gzip_base64(self, m_fn, DS):
        data = VMW_METADATA_YAML.encode("utf-8")
        data = gzip.compress(data)
        data = base64.b64encode(data)
        m_fn.side_effect = [data, "gzip+base64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_get_data_metadata_gz_b64(self, m_fn, DS):
        data = VMW_METADATA_YAML.encode("utf-8")
        data = gzip.compress(data)
        data = base64.b64encode(data)
        m_fn.side_effect = [data, "gz+b64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_metadata_single_ssh_key(self, m_fn, DS):
        metadata = DataSourceVMware.load_json_or_yaml(VMW_METADATA_YAML)
        metadata["public_keys"] = VMW_SINGLE_KEY
        metadata_yaml = safeyaml.dumps(metadata)
        m_fn.side_effect = [metadata_yaml, "", "", ""]
        self.assert_metadata(DS, metadata, m_fn, m_fn_call_count=4)

    @mock.patch(
        "cloudinit.sources.DataSourceVMware.guestinfo_envvar_get_value"
    )
    def test_metadata_multiple_ssh_keys(self, m_fn, DS):
        metadata = DataSourceVMware.load_json_or_yaml(VMW_METADATA_YAML)
        metadata["public_keys"] = VMW_MULTIPLE_KEYS
        metadata_yaml = safeyaml.dumps(metadata)
        m_fn.side_effect = [metadata_yaml, "", "", ""]
        self.assert_metadata(DS, metadata, m_fn, m_fn_call_count=4)


class TestDataSourceVMwareGuestInfo:
    """
    Test the guestinfo transport on a VMware platform.
    """

    @pytest.fixture(autouse=True)
    def create_files(self, fake_filesystem, tmpdir):
        populate_dir(
            str(tmpdir),
            {
                DataSourceVMware.PRODUCT_UUID_FILE_PATH: PRODUCT_UUID,
                PRODUCT_NAME_FILE_PATH: PRODUCT_NAME,
            },
        )

    def assert_get_data_ok(self, DS, m_fn, m_fn_call_count=6):
        ds = DS(settings.CFG_BUILTIN)
        ret = ds.get_data()
        assert ret
        assert m_fn_call_count == m_fn.call_count
        assert (
            ds.data_access_method
            == DataSourceVMware.DATA_ACCESS_METHOD_GUESTINFO
        )
        return ds

    def assert_metadata(self, DS, metadata, m_fn, m_fn_call_count=6):
        ds = self.assert_get_data_ok(DS, m_fn, m_fn_call_count)
        assert_metadata(ds, metadata)

    def test_ds_valid_on_vmware_platform(self):
        system_type = dmi.read_dmi_data("system-product-name")
        assert system_type == PRODUCT_NAME

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_subplatform(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        ds = self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)
        assert ds.subplatform == "%s (%s)" % (
            DataSourceVMware.DATA_ACCESS_METHOD_GUESTINFO,
            DataSourceVMware.get_guestinfo_key_name("metadata"),
        )

        # Test to ensure that network is configured from metadata on each boot.
        assert (
            DataSourceVMware.DEFAULT_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.default_update_events[EventScope.NETWORK]
        )
        assert (
            DataSourceVMware.SUPPORTED_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.supported_update_events[EventScope.NETWORK]
        )

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_with_vmware_rpctool(
        self, m_which_fn, m_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.exec_vmware_rpctool")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_non_zero_exit_code_fallback_to_vmtoolsd(
        self, m_which_fn, m_exec_vmware_rpctool_fn, m_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_exec_vmware_rpctool_fn.side_effect = ProcessExecutionError(
            exit_code=1
        )
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.exec_vmware_rpctool")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_vmware_rpctool_not_found_fallback_to_vmtoolsd(
        self, m_which_fn, m_exec_vmware_rpctool_fn, m_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", None]
        m_fn.side_effect = [VMW_METADATA_YAML, "", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_userdata_only(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_fn.side_effect = ["", VMW_USERDATA_YAML, "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_vendordata_only(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_fn.side_effect = ["", "", VMW_VENDORDATA_YAML, ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_metadata_single_ssh_key(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        metadata = DataSourceVMware.load_json_or_yaml(VMW_METADATA_YAML)
        metadata["public_keys"] = VMW_SINGLE_KEY
        metadata_yaml = safeyaml.dumps(metadata)
        m_fn.side_effect = [metadata_yaml, "", "", ""]
        self.assert_metadata(DS, metadata, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_metadata_multiple_ssh_keys(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        metadata = DataSourceVMware.load_json_or_yaml(VMW_METADATA_YAML)
        metadata["public_keys"] = VMW_MULTIPLE_KEYS
        metadata_yaml = safeyaml.dumps(metadata)
        m_fn.side_effect = [metadata_yaml, "", "", ""]
        self.assert_metadata(DS, metadata, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_base64(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        data = base64.b64encode(VMW_METADATA_YAML.encode("utf-8"))
        m_fn.side_effect = [data, "base64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_b64(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        data = base64.b64encode(VMW_METADATA_YAML.encode("utf-8"))
        m_fn.side_effect = [data, "b64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_gzip_base64(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        data = VMW_METADATA_YAML.encode("utf-8")
        data = gzip.compress(data)
        data = base64.b64encode(data)
        m_fn.side_effect = [data, "gzip+base64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_get_data_metadata_gz_b64(self, m_which_fn, m_fn, DS):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        data = VMW_METADATA_YAML.encode("utf-8")
        data = gzip.compress(data)
        data = base64.b64encode(data)
        m_fn.side_effect = [data, "gz+b64", "", ""]
        self.assert_get_data_ok(DS, m_fn, m_fn_call_count=4)

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_set_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_advertise_update_events(
        self, m_which_fn, m_get_fn, m_set_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_get_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        ds = self.assert_get_data_ok(DS, m_get_fn, m_fn_call_count=4)
        supported_events, enabled_events = ds.advertise_update_events({})
        assert 2 == m_set_fn.call_count
        assert "network=boot;boot-new-instance;hotplug" == supported_events
        assert "network=boot-new-instance;hotplug" == enabled_events

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_set_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_advertise_update_events_with_events_from_user_data(
        self, m_which_fn, m_get_fn, m_set_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_get_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        ds = self.assert_get_data_ok(DS, m_get_fn, m_fn_call_count=4)
        supported_events, enabled_events = ds.advertise_update_events(
            {
                "updates": {
                    "network": {
                        "when": ["boot"],
                    },
                },
            }
        )
        assert 2 == m_set_fn.call_count
        assert "network=boot;boot-new-instance;hotplug" == supported_events
        assert "network=boot" == enabled_events

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    @mock.patch("cloudinit.sources.DataSourceVMware.which")
    def test_extra_hotplug_udev_rules_with_net_drivers(
        self, m_which_fn, m_get_fn, DS, tmpdir
    ):
        m_which_fn.side_effect = ["vmtoolsd", "vmware-rpctool"]
        m_get_fn.side_effect = [
            VMW_METADATA_YAML_WITH_NET_DRIVERS,
            "",
            "",
            "",
            "",
            "",
        ]
        ds = self.assert_get_data_ok(DS, m_get_fn, m_fn_call_count=4)
        ds.init_extra_hotplug_udev_rules()

        assert (
            VMW_EXPECTED_EXTRA_HOTPLUG_UDEV_RULES_VMXNET
            == ds.extra_hotplug_udev_rules
        )


class TestDataSourceVMwareGuestInfo_InvalidPlatform:
    """
    Test the guestinfo transport on a non-VMware platform.
    """

    @pytest.fixture(autouse=True)
    def create_files(self, fake_filesystem, tmpdir):
        populate_dir(
            str(tmpdir),
            {DataSourceVMware.PRODUCT_UUID_FILE_PATH: PRODUCT_UUID},
        )

    @mock.patch("cloudinit.sources.DataSourceVMware.guestinfo_get_value")
    def test_ds_invalid_on_non_vmware_platform(self, m_fn, DS):
        system_type = dmi.read_dmi_data("system-product-name")
        assert system_type is None

        m_fn.side_effect = [VMW_METADATA_YAML, "", "", "", "", ""]
        ds = DS(settings.CFG_BUILTIN)
        ret = ds.get_data()
        assert not ret


class TestDataSourceVMwareIMC:
    """
    Test the VMware Guest OS Customization transport
    """

    def test_get_subplatform(self, DS, tmpdir):
        ds = DS({"disable_vmware_customization": True})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)
        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = dedent(
            """\
            {
              "instance-id": "cloud-vm",
              "local-hostname": "my-host.domain.com",
              "network": {
                "version": 2,
                "ethernets": {
                  "eths": {
                    "match": {
                      "name": "ens*"
                    },
                    "dhcp4": true
                  }
                }
              }
            }
            """
        )
        util.write_file(metadata_file, metadata_content)

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds._get_data,
            )
            assert result

        assert ds.subplatform == "%s (%s)" % (
            DataSourceVMware.DATA_ACCESS_METHOD_IMC,
            DataSourceVMware.get_imc_key_name("metadata"),
        )

        # Test to ensure that network is configured from metadata on each boot.
        assert (
            DataSourceVMware.DEFAULT_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.default_update_events[EventScope.NETWORK]
        )
        assert (
            DataSourceVMware.SUPPORTED_UPDATE_EVENTS[EventScope.NETWORK]
            == ds.supported_update_events[EventScope.NETWORK]
        )

    def test_get_data_false_on_none_dmi_data(self, caplog, DS):
        """When dmi for system-product-name is None, get_data returns False."""
        ds = DS({})
        result = wrap_and_call(
            "cloudinit.sources.DataSourceVMware",
            {
                "dmi.read_dmi_data": None,
            },
            ds.get_data,
        )
        assert not result, "Expected False return from ds.get_data"
        assert "No system-product-name found" in caplog.text

    def test_get_imc_data_vmware_customization_disabled(
        self, caplog, DS, tmpdir
    ):
        """
        When vmware customization is disabled via sys_cfg and
        allow_raw_data is disabled via ds_cfg, log a message.
        """
        ds = DS(
            {
                "disable_vmware_customization": True,
                "datasource": {"VMware": {"allow_raw_data": False}},
            },
        )
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [MISC]
            MARKER-ID = 12345345
            """
        )
        util.write_file(conf_file, conf_content)
        result = wrap_and_call(
            "cloudinit.sources.DataSourceVMware",
            {
                "dmi.read_dmi_data": "vmware",
            },
            ds.get_imc_data_fn,
        )
        assert result == (None, None, None)
        assert "Customization for VMware platform is disabled" in caplog.text

    def test_get_imc_data_vmware_customization_sys_cfg_disabled(
        self, caplog, DS, tmpdir
    ):
        """
        When vmware customization is disabled via sys_cfg and
        no meta data is found, log a message.
        """
        ds = DS(
            {
                "disable_vmware_customization": True,
                "datasource": {"VMware": {"allow_raw_data": True}},
            },
        )
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [MISC]
            MARKER-ID = 12345345
            """
        )
        util.write_file(conf_file, conf_content)
        result = wrap_and_call(
            "cloudinit.sources.DataSourceVMware",
            {
                "dmi.read_dmi_data": "vmware",
                "util.del_dir": True,
                "guestcust_util.search_file": tmpdir,
                "guestcust_util.wait_for_cust_cfg_file": conf_file,
            },
            ds.get_imc_data_fn,
        )
        assert result == (None, None, None)
        assert (
            "No allowed customization configuration data found" in caplog.text
        )

    def test_get_imc_data_allow_raw_data_disabled(self, caplog, DS, tmpdir):
        """
        When allow_raw_data is disabled via ds_cfg and
        meta data is found, log a message.
        """
        ds = DS(
            {
                "disable_vmware_customization": False,
                "datasource": {"VMware": {"allow_raw_data": False}},
            },
        )

        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)
        result = wrap_and_call(
            "cloudinit.sources.DataSourceVMware",
            {
                "dmi.read_dmi_data": "vmware",
                "util.del_dir": True,
                "guestcust_util.search_file": tmpdir,
                "guestcust_util.wait_for_cust_cfg_file": conf_file,
            },
            ds.get_imc_data_fn,
        )
        assert result == (None, None, None)
        assert (
            "No allowed customization configuration data found" in caplog.text
        )

    @pytest.mark.allow_subp_for("vmware-rpctool")
    def test_get_imc_data_vmware_customization_enabled(
        self, caplog, DS, tmpdir
    ):
        """
        When cloud-init workflow for vmware is enabled via sys_cfg log a
        message.
        """
        ds = DS({"disable_vmware_customization": False})
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CUSTOM-SCRIPT]
            SCRIPT-NAME = test-script
            [MISC]
            MARKER-ID = 12345345
            """
        )
        util.write_file(conf_file, conf_content)
        with mock.patch(
            MPATH + "guestcust_util.get_tools_config",
            return_value="true",
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                },
                ds.get_imc_data_fn,
            )
            assert result == (None, None, None)
        custom_script = os.path.join(tmpdir, "test-script")
        assert "Script %s not found!!" % custom_script in caplog.text

    def test_get_imc_data_cust_script_disabled(self, caplog, DS, tmpdir):
        """
        If custom script is disabled by VMware tools configuration,
        log a message.
        """
        ds = DS({"disable_vmware_customization": False})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CUSTOM-SCRIPT]
            SCRIPT-NAME = test-script
            [MISC]
            MARKER-ID = 12345346
            """
        )
        util.write_file(conf_file, conf_content)
        # Prepare the custom script
        customscript = os.path.join(tmpdir, "test-script")
        util.write_file(customscript, "This is the post cust script")

        with mock.patch(
            MPATH + "guestcust_util.get_tools_config",
            return_value="invalid",
        ):
            with mock.patch(
                MPATH + "guestcust_util.set_customization_status",
                return_value=("msg", b""),
            ):
                result = wrap_and_call(
                    "cloudinit.sources.DataSourceVMware",
                    {
                        "dmi.read_dmi_data": "vmware",
                        "util.del_dir": True,
                        "guestcust_util.search_file": tmpdir,
                        "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    },
                    ds.get_imc_data_fn,
                )
                assert result == (None, None, None)
        assert "Custom script is disabled by VM Administrator" in caplog.text

    def test_get_imc_data_cust_script_enabled(self, caplog, DS, tmpdir):
        """
        If custom script is enabled by VMware tools configuration,
        execute the script.
        """
        ds = DS({"disable_vmware_customization": False})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CUSTOM-SCRIPT]
            SCRIPT-NAME = test-script
            [MISC]
            MARKER-ID = 12345346
            """
        )
        util.write_file(conf_file, conf_content)

        # Mock custom script is enabled by return true when calling
        # get_tools_config
        with mock.patch(
            MPATH + "guestcust_util.get_tools_config",
            return_value="true",
        ):
            with mock.patch(
                MPATH + "guestcust_util.set_customization_status",
                return_value=("msg", b""),
            ):
                result = wrap_and_call(
                    "cloudinit.sources.DataSourceVMware",
                    {
                        "dmi.read_dmi_data": "vmware",
                        "util.del_dir": True,
                        "guestcust_util.search_file": tmpdir,
                        "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    },
                    ds.get_imc_data_fn,
                )
                assert result == (None, None, None)
        # Verify custom script is trying to be executed
        custom_script = os.path.join(tmpdir, "test-script")
        assert "Script %s not found!!" % custom_script in caplog.text

    def test_get_imc_data_force_run_post_script_is_yes(
        self, caplog, DS, tmpdir
    ):
        """
        If DEFAULT-RUN-POST-CUST-SCRIPT is yes, custom script could run if
        enable-custom-scripts is not defined in VM Tools configuration
        """
        ds = DS({"disable_vmware_customization": False})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        # set DEFAULT-RUN-POST-CUST-SCRIPT = yes so that enable-custom-scripts
        # default value is TRUE
        conf_content = dedent(
            """\
            [CUSTOM-SCRIPT]
            SCRIPT-NAME = test-script
            [MISC]
            MARKER-ID = 12345346
            DEFAULT-RUN-POST-CUST-SCRIPT = yes
            """
        )
        util.write_file(conf_file, conf_content)

        # Mock get_tools_config(section, key, defaultVal) to return
        # defaultVal
        def my_get_tools_config(*args, **kwargs):
            return args[2]

        with mock.patch(
            MPATH + "guestcust_util.get_tools_config",
            side_effect=my_get_tools_config,
        ):
            with mock.patch(
                MPATH + "guestcust_util.set_customization_status",
                return_value=("msg", b""),
            ):
                result = wrap_and_call(
                    "cloudinit.sources.DataSourceVMware",
                    {
                        "dmi.read_dmi_data": "vmware",
                        "util.del_dir": True,
                        "guestcust_util.search_file": tmpdir,
                        "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    },
                    ds.get_imc_data_fn,
                )
                assert result == (None, None, None)
        # Verify custom script still runs although it is
        # disabled by VMware Tools
        custom_script = os.path.join(tmpdir, "test-script")
        assert "Script %s not found!!" % custom_script in caplog.text

    def test_get_data_cloudinit_metadata_json(self, DS, tmpdir):
        """
        Test metadata can be loaded to cloud-init metadata and network.
        The metadata format is json.
        """
        ds = DS({"disable_vmware_customization": True})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)
        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = dedent(
            """\
            {
              "instance-id": "cloud-vm",
              "local-hostname": "my-host.domain.com",
              "network": {
                "version": 2,
                "ethernets": {
                  "eths": {
                    "match": {
                      "name": "ens*"
                    },
                    "dhcp4": true
                  }
                }
              }
            }
            """
        )
        util.write_file(metadata_file, metadata_content)

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds._get_data,
            )
            assert result
        assert "cloud-vm" == ds.metadata["instance-id"]
        assert "my-host.domain.com" == ds.metadata["local-hostname"]
        assert 2 == ds.network_config["version"]
        assert ds.network_config["ethernets"]["eths"]["dhcp4"]

    def test_get_data_cloudinit_metadata_yaml(self, DS, tmpdir):
        """
        Test metadata can be loaded to cloud-init metadata and network.
        The metadata format is yaml.
        """
        ds = DS({"disable_vmware_customization": True})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)
        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = dedent(
            """\
            instance-id: cloud-vm
            local-hostname: my-host.domain.com
            network:
                version: 2
                ethernets:
                    nics:
                        match:
                            name: ens*
                        dhcp4: yes
            """
        )
        util.write_file(metadata_file, metadata_content)

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds._get_data,
            )
            assert result
        assert "cloud-vm" == ds.metadata["instance-id"]
        assert "my-host.domain.com" == ds.metadata["local-hostname"]
        assert 2 == ds.network_config["version"]
        assert ds.network_config["ethernets"]["nics"]["dhcp4"]

    def test_get_imc_data_cloudinit_metadata_not_valid(
        self, caplog, DS, tmpdir
    ):
        """
        Test metadata is not JSON or YAML format, log a message
        """
        ds = DS({"disable_vmware_customization": True})

        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)

        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = "[This is not json or yaml format]a=b"
        util.write_file(metadata_file, metadata_content)

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds.get_data,
            )
        assert not result
        assert (
            "expected '<document start>', but found '<scalar>'" in caplog.text
        )

    def test_get_imc_data_cloudinit_metadata_not_found(
        self, caplog, DS, tmpdir
    ):
        """
        Test metadata file can't be found, log a message
        """
        ds = DS({"disable_vmware_customization": True})
        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            """
        )
        util.write_file(conf_file, conf_content)
        # Don't prepare the meta data file

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds.get_imc_data_fn,
            )
            assert result == (None, None, None)
        assert "Meta data file is not found" in caplog.text

    def test_get_data_cloudinit_userdata(self, DS, tmpdir):
        """
        Test user data can be loaded to cloud-init user data.
        """
        ds = DS({"disable_vmware_customization": False})

        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            USERDATA = test-user
            """
        )
        util.write_file(conf_file, conf_content)

        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = dedent(
            """\
            instance-id: cloud-vm
            local-hostname: my-host.domain.com
            network:
                version: 2
                ethernets:
                    nics:
                        match:
                            name: ens*
                        dhcp4: yes
            """
        )
        util.write_file(metadata_file, metadata_content)

        # Prepare the user data file
        userdata_file = os.path.join(tmpdir, "test-user")
        userdata_content = "This is the user data"
        util.write_file(userdata_file, userdata_content)

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds._get_data,
            )
            assert result
        assert "cloud-vm" == ds.metadata["instance-id"]
        assert userdata_content == ds.userdata_raw

    def test_get_imc_data_cloudinit_userdata_not_found(
        self, caplog, DS, tmpdir
    ):
        """
        Test userdata file can't be found.
        """
        ds = DS({"disable_vmware_customization": True})

        # Prepare the conf file
        conf_file = os.path.join(tmpdir, "test-cust")
        conf_content = dedent(
            """\
            [CLOUDINIT]
            METADATA = test-meta
            USERDATA = test-user
            """
        )
        util.write_file(conf_file, conf_content)

        # Prepare the meta data file
        metadata_file = os.path.join(tmpdir, "test-meta")
        metadata_content = dedent(
            """\
            instance-id: cloud-vm
            local-hostname: my-host.domain.com
            network:
                version: 2
                ethernets:
                    nics:
                        match:
                            name: ens*
                        dhcp4: yes
            """
        )
        util.write_file(metadata_file, metadata_content)

        # Don't prepare the user data file

        with mock.patch(
            MPATH + "guestcust_util.set_customization_status",
            return_value=("msg", b""),
        ):
            result = wrap_and_call(
                "cloudinit.sources.DataSourceVMware",
                {
                    "dmi.read_dmi_data": "vmware",
                    "util.del_dir": True,
                    "guestcust_util.search_file": tmpdir,
                    "guestcust_util.wait_for_cust_cfg_file": conf_file,
                    "guestcust_util.get_imc_dir_path": tmpdir,
                },
                ds.get_imc_data_fn,
            )
            assert result == (None, None, None)
        assert "Userdata file is not found" in caplog.text


class TestDataSourceVMwareIMC_MarkerFiles:

    def test_false_when_markerid_none(self, tmpdir):
        """Return False when markerid provided is None."""
        assert not guestcust_util.check_marker_exists(
            markerid=None, marker_dir=tmpdir
        )

    def test_markerid_file_exist(self, tmpdir):
        """Return False when markerid file path does not exist,
        True otherwise."""
        assert not guestcust_util.check_marker_exists("123", tmpdir)
        marker_file = os.path.join(tmpdir, ".markerfile-123.txt")
        util.write_file(marker_file, "")
        assert guestcust_util.check_marker_exists("123", tmpdir)

    def test_marker_file_setup(self, tmpdir):
        """Test creation of marker files."""
        markerfilepath = os.path.join(tmpdir, ".markerfile-hi.txt")
        assert not os.path.exists(markerfilepath)
        guestcust_util.setup_marker_files(marker_id="hi", marker_dir=tmpdir)
        assert os.path.exists(markerfilepath)


def assert_metadata(ds, metadata):
    assert metadata.get("instance-id") == ds.get_instance_id()
    assert metadata.get("local-hostname") == ds.get_hostname().hostname

    expected_public_keys = metadata.get("public_keys")
    if not isinstance(expected_public_keys, list):
        expected_public_keys = [expected_public_keys]

    assert expected_public_keys == ds.get_public_ssh_keys()
    assert isinstance(ds.get_public_ssh_keys(), list)


@pytest.fixture
def DS(paths):
    return lambda sys_cfg: DataSourceVMware.DataSourceVMware(
        sys_cfg, {}, paths
    )
